from typing import Any, Optional, Type

from pydantic import BaseModel, validator

from phidata.utils.log import logger


class InfraResource(BaseModel):
    """Base class for all Phidata infrastructure resources.
    All Models in the phidata.infra.*.resource modules are expected to be subclasses of this Model.
    The rationale for using a pydantic model here is that the data which populates this
    model comes from an external infrastructure api, which can return anything so we need to
    validate the data were ingesting.
    """

    # type of resource
    resource_type: Optional[str] = None
    # name of resource
    name: Optional[str] = None

    # resource management
    # If True, skip resource creation if active resources with the same name exist.
    use_cache: bool = True
    # If True, logs extra debug messages
    use_verbose_logs: bool = False

    # If True, marks skip_create, skip_delete, skip_update to True
    read_only: bool = False
    # If True, does not create the resource when `phi ws up` command is run
    skip_create: bool = False
    # If True, does not delete the resource when `phi ws down` command is run
    skip_delete: bool = False
    # If True, does not update the resource when `phi ws patch` command is run
    skip_update: bool = False

    # If True, waits for the resource to be created
    wait_for_creation: bool = True
    # If True, waits for the resource to be deleted
    wait_for_deletion: bool = True
    # The amount of time in seconds to wait between attempts.
    waiter_delay: int = 30
    # The maximum number of attempts to be made.
    waiter_max_attempts: int = 50

    active_resource: Optional[Any] = None
    active_resource_class: Optional[Type] = None

    def get_resource_name(self) -> Optional[str]:
        return self.name

    def get_resource_type(self) -> Optional[str]:
        return self.resource_type

    def verbose_log(self, msg: Any) -> None:
        if self.use_verbose_logs:
            logger.debug(msg)

    @validator("skip_create", pre=True, always=True)
    def set_skip_create(cls, skip_create, values):
        read_only = values.get("read_only", False)
        return True if read_only else skip_create

    @validator("skip_delete", pre=True, always=True)
    def set_skip_delete(cls, skip_delete, values):
        read_only = values.get("read_only", False)
        return True if read_only else skip_delete

    @validator("skip_update", pre=True, always=True)
    def set_skip_update(cls, skip_update, values):
        read_only = values.get("read_only", False)
        return True if read_only else skip_update

    """
    ## Functions to be implemented by subclasses
    def create(self, infra_api_client: InfraApiClient) -> bool:
    def read(self, infra_api_client: InfraApiClient) -> bool:
    def update(self, infra_api_client: InfraApiClient) -> bool:
    def delete(self, infra_api_client: InfraApiClient) -> bool:
    def is_active(self, infra_api_client: InfraApiClient) -> bool:
    """

    class Config:
        # https://pydantic-docs.helpmanual.io/usage/model_config/
        # If we need to use an alias for fields of subclasses, eg: Kubeconfig
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        use_enum_values = True



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, Any, List, Union
from typing_extensions import Literal

from phidata.app import PhidataApp, PhidataAppArgs
from phidata.infra.docker.resource.network import DockerNetwork
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.k8s.create.apps.v1.deployment import CreateDeployment
from phidata.infra.k8s.create.core.v1.secret import CreateSecret
from phidata.infra.k8s.create.core.v1.config_map import CreateConfigMap
from phidata.infra.k8s.create.core.v1.container import CreateContainer
from phidata.infra.k8s.create.core.v1.volume import (
    CreateVolume,
    EmptyDirVolumeSource,
    VolumeType,
)
from phidata.infra.k8s.create.common.port import CreatePort
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.common import (
    get_image_str,
    get_default_container_name,
    get_default_configmap_name,
    get_default_secret_name,
    get_default_service_name,
    get_default_volume_name,
    get_default_deploy_name,
    get_default_pod_name,
)
from phidata.utils.cli_console import print_error
from phidata.utils.log import logger


class AirflowArgs(PhidataAppArgs):
    name: str = "airflow"
    version: str = "1"
    enabled: bool = True

    # Webserver Args
    ws_enabled: bool = True
    ws_image_name: str = "phidata/airflow"
    ws_image_tag: str = "1"
    ws_entrypoint: str = "/entrypoint.sh"
    ws_command: str = "webserver"
    ws_env: Optional[Dict[str, str]] = None
    # Set as True to init the airflow db before running the webserver
    ws_init_airflow_db: bool = False
    # Set as True to wait for the airflow db to be available before running the webserver
    ws_wait_for_db: bool = False
    # Set as True to wait for redis to be available before running the webserver
    ws_wait_for_redis: bool = False
    # Webserver Docker Args
    ws_container_name: str = "airflow-ws"
    ws_container_port: str = "8080"
    ws_container_host_port: int = 8080
    # Webserver K8s Args
    # ws_rg_name: str = "airflow-webserver"
    # ws_svc_name: str = "airflow-webserver-svc"
    # ws_deploy_name: str = "airflow-webserver-deploy"
    # ws_replicas: int = 1
    # ws_port_name: str = "http"
    # ws_port_number: int = 8080

    # Scheduler Args
    scheduler_enabled: bool = True
    scheduler_image_name: str = "phidata/airflow"
    scheduler_image_tag: str = "0.1"
    scheduler_entrypoint: str = "/entrypoint.sh"
    scheduler_command: str = "scheduler"
    scheduler_env: Optional[Dict[str, str]] = None
    scheduler_init_airflow_db: bool = False
    scheduler_wait_for_db: bool = False
    scheduler_wait_for_redis: bool = False
    scheduler_container_name: str = "airflow-scheduler"

    # Worker Args
    worker_enabled: bool = True
    worker_image_name: str = "phidata/airflow"
    worker_image_tag: str = "0.1"
    worker_entrypoint: str = "/entrypoint.sh"
    worker_command: str = "worker"
    worker_env: Optional[Dict[str, str]] = None
    worker_container_name: str = "airflow-worker"

    # Flower Args
    flower_enabled: bool = True
    flower_image_name: str = "phidata/airflow"
    flower_image_tag: str = "0.1"
    flower_entrypoint: str = "/entrypoint.sh"
    flower_command: str = "flower"
    flower_env: Optional[Dict[str, str]] = None
    flower_container_name: str = "airflow-flower"
    flower_container_port: str = "5555"
    flower_container_host_port: int = 5555

    # DB Args
    db_enabled: bool = True
    db_type: Literal["postgres", "mysql"] = "postgres"
    # Postgres as DB args
    pg_image_name: str = "postgres"
    pg_image_tag: str = "13"
    pg_driver: str = "postgresql"
    pg_env: Optional[Dict[str, str]] = None
    pg_user: str = "airflow"
    pg_password: str = "airflow"
    pg_schema: str = "airflow"
    pg_extras: str = ""
    pg_data_path: str = "/var/lib/postgresql/data/"
    # Postgres Docker Args
    pg_container_name: str = "airflow-pg"
    pg_container_port: str = "5432"
    pg_container_host_port: int = 5432
    # We currently let docker manage the storage of our database data
    pg_volume_name: str = "airflow_pg_data"
    # The directory on the users (host) machine containing the postgres data files
    # pg_data_host_path: Optional[str] = None
    # K8s Args
    # pg_rg_name: str = "airflow-pg"
    # pg_svc_name: str = "airflow-pg-svc"
    # pg_deploy_name: str = "airflow-pg-deploy"
    # pg_replicas: int = 1
    # pg_pvc_name: str = "airflow-pg-pvc"
    # pg_cm_name: str = "airflow-pg-cm"

    # Redis Args
    redis_enabled: bool = True
    redis_image_name: str = "redis"
    redis_image_tag: str = "latest"
    redis_driver: str = "redis"
    redis_user: str = "airflow"
    redis_pass: str = "airflow"
    redis_schema: str = "1"
    # Redis Docker Args
    redis_container_name: str = "airflow-redis"
    redis_container_port: str = "6379"
    redis_container_host_port: int = 6379
    # K8s Args
    # redis_svc_name: str = "airflow-redis-svc"
    # redis_deploy_name: str = "airflow-redis-deploy"

    # Common Args
    # The directory on the users (host) machine containing the airflow DAG files, starting from the workspace_root_path.
    airflow_dags_host_dir_path: str = "/projects"
    # The path on the users (host) machine for the dir of airflow.cfg file, starting from the workspace_root_path.
    airflow_conf_host_dir_path: str = "/workspace/airflow"
    # The directory in the container containing the airflow pipelines
    airflow__core__dags_folder: str = "/usr/local/airflow/dags"
    # The path in the container for the airflow.cfg file
    airflow__core__conf_dir: str = "/usr/local/airflow"

    # Common K8s Args
    # cm_name: str = "airflow-cm"
    # Storage Classes
    # ssd_rg_name: str = "airflow-ssd"
    # ssd_storage_class_name: str = "airflow-ssd"


class Airflow(PhidataApp):
    def __init__(
        self,
        name: str = "airflow",
        version: str = "1",
        enabled: bool = True,
        ws_enabled: bool = False,
        ws_image_name: str = "phidata/airflow",
        ws_image_tag: str = "0.5",
        ws_entrypoint: str = "/entrypoint.sh",
        ws_command: str = "webserver",
        ws_env: Optional[Dict[str, str]] = None,
        ws_init_airflow_db: bool = False,
        ws_wait_for_db: bool = True,
        ws_wait_for_redis: bool = True,
        ws_container_name: str = "airflow-ws",
        ws_container_port: str = "8080",
        ws_container_host_port: int = 9090,
        scheduler_enabled: bool = False,
        scheduler_image_name: str = "phidata/airflow",
        scheduler_image_tag: str = "0.5",
        scheduler_entrypoint: str = "/entrypoint.sh",
        scheduler_command: str = "scheduler",
        scheduler_env: Optional[Dict[str, str]] = None,
        scheduler_init_airflow_db: bool = True,
        scheduler_wait_for_db: bool = True,
        scheduler_wait_for_redis: bool = True,
        scheduler_container_name: str = "airflow-scheduler",
        worker_enabled: bool = False,
        worker_image_name: str = "phidata/airflow",
        worker_image_tag: str = "0.1",
        worker_entrypoint: str = "/entrypoint.sh",
        worker_command: str = "worker",
        worker_env: Optional[Dict[str, str]] = None,
        worker_container_name: str = "airflow-worker",
        flower_enabled: bool = False,
        flower_image_name: str = "phidata/airflow",
        flower_image_tag: str = "0.1",
        flower_entrypoint: str = "/entrypoint.sh",
        flower_command: str = "flower",
        flower_env: Optional[Dict[str, str]] = None,
        flower_container_name: str = "airflow-flower",
        flower_container_port: str = "5555",
        flower_container_host_port: int = 6000,
        db_enabled: bool = False,
        db_type: Literal["postgres", "mysql"] = "postgres",
        pg_image_name: str = "postgres",
        pg_image_tag: str = "13",
        pg_driver: str = "postgresql",
        pg_env: Optional[Dict[str, str]] = None,
        pg_user: str = "airflow",
        pg_password: str = "airflow",
        pg_schema: str = "airflow",
        pg_data_path: str = "/var/lib/postgresql/data/",
        pg_container_name: str = "airflow-pg",
        pg_container_port: str = "5432",
        pg_container_host_port: int = 5500,
        pg_volume_name: str = "airflow_pg_data",
        redis_enabled: bool = False,
        redis_image_name: str = "redis",
        redis_image_tag: str = "latest",
        redis_driver: str = "redis",
        redis_user: str = "airflow",
        redis_pass: str = "airflow",
        redis_schema: str = "1",
        redis_container_name: str = "airflow-redis",
        redis_container_port: str = "6379",
        redis_container_host_port: int = 6500,
        airflow_dags_host_dir_path: str = "/projects",
        airflow_conf_host_dir_path: str = "/workspace/airflow",
        airflow__core__dags_folder: str = "/usr/local/airflow/dags",
        airflow__core__conf_dir: str = "/usr/local/airflow",
    ):

        super().__init__()
        try:
            self.args: AirflowArgs = AirflowArgs(
                name=name,
                version=version,
                enabled=enabled,
                ws_enabled=ws_enabled,
                ws_image_name=ws_image_name,
                ws_image_tag=ws_image_tag,
                ws_entrypoint=ws_entrypoint,
                ws_command=ws_command,
                ws_env=ws_env,
                ws_init_airflow_db=ws_init_airflow_db,
                ws_wait_for_db=ws_wait_for_db,
                ws_wait_for_redis=ws_wait_for_redis,
                ws_container_name=ws_container_name,
                ws_container_port=ws_container_port,
                ws_container_host_port=ws_container_host_port,
                scheduler_enabled=scheduler_enabled,
                scheduler_image_name=scheduler_image_name,
                scheduler_image_tag=scheduler_image_tag,
                scheduler_entrypoint=scheduler_entrypoint,
                scheduler_command=scheduler_command,
                scheduler_env=scheduler_env,
                scheduler_init_airflow_db=scheduler_init_airflow_db,
                scheduler_wait_for_db=scheduler_wait_for_db,
                scheduler_wait_for_redis=scheduler_wait_for_redis,
                scheduler_container_name=scheduler_container_name,
                worker_enabled=worker_enabled,
                worker_image_name=worker_image_name,
                worker_image_tag=worker_image_tag,
                worker_entrypoint=worker_entrypoint,
                worker_command=worker_command,
                worker_env=worker_env,
                worker_container_name=worker_container_name,
                flower_enabled=flower_enabled,
                flower_image_name=flower_image_name,
                flower_image_tag=flower_image_tag,
                flower_entrypoint=flower_entrypoint,
                flower_command=flower_command,
                flower_env=flower_env,
                flower_container_name=flower_container_name,
                flower_container_port=flower_container_port,
                flower_container_host_port=flower_container_host_port,
                db_enabled=db_enabled,
                db_type=db_type,
                pg_image_name=pg_image_name,
                pg_image_tag=pg_image_tag,
                pg_driver=pg_driver,
                pg_env=pg_env,
                pg_user=pg_user,
                pg_password=pg_password,
                pg_schema=pg_schema,
                pg_data_path=pg_data_path,
                pg_container_name=pg_container_name,
                pg_container_port=pg_container_port,
                pg_container_host_port=pg_container_host_port,
                pg_volume_name=pg_volume_name,
                redis_enabled=redis_enabled,
                redis_image_name=redis_image_name,
                redis_image_tag=redis_image_tag,
                redis_driver=redis_driver,
                redis_user=redis_user,
                redis_pass=redis_pass,
                redis_schema=redis_schema,
                redis_container_name=redis_container_name,
                redis_container_port=redis_container_port,
                redis_container_host_port=redis_container_host_port,
                airflow_dags_host_dir_path=airflow_dags_host_dir_path,
                airflow_conf_host_dir_path=airflow_conf_host_dir_path,
                airflow__core__dags_folder=airflow__core__dags_folder,
                airflow__core__conf_dir=airflow__core__conf_dir,
            )
        except Exception as e:
            # from phidata.utils.cli_console import print_validation_errors
            #
            # print_validation_errors(e.args)
            raise

        # # these variables can now be accessed using the object instance
        # # Assign the values from self.args so pydantic handles type conversion
        # self.name: str = self.args.name
        # self.version: str = self.args.version
        # self.enabled: bool = self.args.enabled
        # self.ws_enabled: bool = self.args.ws_enabled
        # self.ws_image_name: str = self.args.ws_image_name
        # self.ws_image_tag: str = self.args.ws_image_tag
        # self.ws_entrypoint: str = self.args.ws_entrypoint
        # self.ws_command: str = self.args.ws_command
        # self.ws_env: Optional[Dict[str, str]] = self.args.ws_env
        # self.ws_init_airflow_db: bool = self.args.ws_init_airflow_db
        # self.ws_wait_for_db: bool = self.args.ws_wait_for_db
        # self.ws_wait_for_redis: bool = self.args.ws_wait_for_redis
        # self.ws_container_name: str = self.args.ws_container_name
        # self.ws_container_port: str = self.args.ws_container_port
        # self.ws_container_host_port: int = self.args.ws_container_host_port
        # self.scheduler_enabled: bool = self.args.scheduler_enabled
        # self.scheduler_image_name: str = self.args.scheduler_image_name
        # self.scheduler_image_tag: str = self.args.scheduler_image_tag
        # self.scheduler_entrypoint: str = self.args.scheduler_entrypoint
        # self.scheduler_command: str = self.args.scheduler_command
        # self.scheduler_env: Optional[Dict[str, str]] = self.args.scheduler_env
        # self.scheduler_init_airflow_db: bool = self.args.scheduler_init_airflow_db
        # self.scheduler_wait_for_db: bool = self.args.scheduler_wait_for_db
        # self.scheduler_wait_for_redis: bool = self.args.scheduler_wait_for_redis
        # self.scheduler_container_name: str = self.args.scheduler_container_name
        # self.worker_enabled: bool = self.args.worker_enabled
        # self.worker_image_name: str = self.args.worker_image_name
        # self.worker_image_tag: str = self.args.worker_image_tag
        # self.worker_entrypoint: str = self.args.worker_entrypoint
        # self.worker_command: str = self.args.worker_command
        # self.worker_env: Optional[Dict[str, str]] = self.args.worker_env
        # self.worker_container_name: str = self.args.worker_container_name
        # self.flower_enabled: bool = self.args.flower_enabled
        # self.flower_image_name: str = self.args.flower_image_name
        # self.flower_image_tag: str = self.args.flower_image_tag
        # self.flower_entrypoint: str = self.args.flower_entrypoint
        # self.flower_command: str = self.args.flower_command
        # self.flower_env: Optional[Dict[str, str]] = self.args.flower_env
        # self.flower_container_name: str = self.args.flower_container_name
        # self.flower_container_port: str = self.args.flower_container_port
        # self.flower_container_host_port: int = self.args.flower_container_host_port
        # self.db_enabled: bool = self.args.db_enabled
        # self.db_type: Literal["postgres", "mysql"] = self.args.db_type
        # self.pg_image_name: str = self.args.pg_image_name
        # self.pg_image_tag: str = self.args.pg_image_tag
        # self.pg_driver: str = self.args.pg_driver
        # self.pg_env: Optional[Dict[str, str]] = self.args.pg_env
        # self.pg_user: str = self.args.pg_user
        # self.pg_password: str = self.args.pg_password
        # self.pg_schema: str = self.args.pg_schema
        # self.pg_data_path: str = self.args.pg_data_path
        # self.pg_container_name: str = self.args.pg_container_name
        # self.pg_container_port: str = self.args.pg_container_port
        # self.pg_container_host_port: int = self.args.pg_container_host_port
        # self.pg_volume_name: str = self.args.pg_volume_name
        # self.redis_enabled: bool = self.args.redis_enabled
        # self.redis_image_name: str = self.args.redis_image_name
        # self.redis_image_tag: str = self.args.redis_image_tag
        # self.redis_driver: str = self.args.redis_driver
        # self.redis_user: str = self.args.redis_user
        # self.redis_pass: str = self.args.redis_pass
        # self.redis_schema: str = self.args.redis_schema
        # self.redis_container_name: str = self.args.redis_container_name
        # self.redis_container_port: str = self.args.redis_container_port
        # self.redis_container_host_port: int = self.args.redis_container_host_port
        # self.airflow_dags_host_dir_path: str = self.args.airflow_dags_host_dir_path
        # self.airflow_conf_host_dir_path: str = self.args.airflow_conf_host_dir_path
        # self.airflow__core__dags_folder: str = self.args.airflow__core__dags_folder
        # self.airflow__core__conf_dir: str = self.args.airflow__core__conf_dir

    def get_postgres_docker_rg(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[DockerResourceGroup]:
        logger.debug(f"Init Postgres DockerResourceGroup")

        _postgres_container_env = {
            "POSTGRES_USER": self.args.pg_user,
            "POSTGRES_PASSWORD": self.args.pg_password,
            "POSTGRES_DB": self.args.pg_schema,
            "PGDATA": self.args.pg_data_path,
        }
        if self.args.pg_env is not None and isinstance(self.args.pg_env, dict):
            _postgres_container_env.update(self.args.pg_env)

        _postgres_container = DockerContainer(
            name=self.args.pg_container_name,
            image="{}:{}".format(self.args.pg_image_name, self.args.pg_image_tag),
            detach=True,
            auto_remove=True,
            remove=True,
            network=docker_build_context.network,
            ports={
                self.args.pg_container_port: self.args.pg_container_host_port,
            },
            volumes={
                self.args.pg_volume_name: {
                    "bind": self.args.pg_data_path,
                    "mode": "rw",
                },
            },
        )

        _postgres_rg = DockerResourceGroup(
            name=self.args.pg_container_name,
            enabled=self.args.db_enabled,
            weight=101,
            containers=[_postgres_container],
        )
        # logger.debug("postgres rg:\n{}".format(_postgres_rg.json(indent=2)))
        return _postgres_rg

    def get_redis_docker_rg(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[DockerResourceGroup]:
        logger.debug(f"Init Redis DockerResourceGroup")

        _redis_container = DockerContainer(
            name=self.args.redis_container_name,
            image="{}:{}".format(self.args.redis_image_name, self.args.redis_image_tag),
            detach=True,
            auto_remove=True,
            remove=True,
            network=docker_build_context.network,
            ports={
                self.args.redis_container_port: self.args.redis_container_host_port,
            },
        )

        _redis_rg = DockerResourceGroup(
            name=self.args.redis_container_name,
            enabled=self.args.redis_enabled,
            weight=101,
            containers=[_redis_container],
        )
        # logger.debug("redis rg:\n{}".format(_redis_rg.json(indent=2)))
        return _redis_rg

    def get_scheduler_docker_rg(
        self,
        docker_build_context: DockerBuildContext,
    ) -> Optional[DockerResourceGroup]:
        logger.debug(f"Init Scheduler DockerResourceGroup")

        _airflow_dags_host_dir_path_str = (
            str(self.args.workspace_dir_path) + self.args.airflow_dags_host_dir_path
        )
        _airflow_conf_host_dir_path = (
            str(self.args.workspace_dir_path) + self.args.airflow_conf_host_dir_path
        )
        _scheduler_container_env = {
            "INIT_AIRFLOW_DB": self.args.scheduler_init_airflow_db,
            "WAIT_FOR_DB": self.args.scheduler_wait_for_db,
            "WAIT_FOR_REDIS": self.args.scheduler_wait_for_redis,
            "AIRFLOW_DB_CONN_URL": self.args.pg_container_name,
            "AIRFLOW_DB_CONN_PORT": self.args.pg_container_port,
            "AIRFLOW_DB_USER": self.args.pg_user,
            "AIRFLOW_DB_PASSWORD": self.args.pg_password,
            "AIRFLOW_SCHEMA": self.args.pg_schema,
            "AIRFLOW_REDIS_CONN_URL": self.args.redis_container_name,
            "AIRFLOW_REDIS_CONN_PORT": self.args.redis_container_port,
            "AIRFLOW_REDIS_USER": self.args.redis_user,
            "AIRFLOW_REDIS_PASSWORD": self.args.redis_pass,
            "AIRFLOW_REDIS_SCHEMA": self.args.redis_schema,
            "AIRFLOW__CORE__EXECUTOR": "CeleryExecutor",
            "AIRFLOW__CORE__SQL_ALCHEMY_CONN": f"{self.args.pg_driver}+psycopg2://{self.args.pg_user}:{self.args.pg_password}@{self.args.pg_container_name}:{self.args.pg_container_port}/{self.args.pg_schema}{self.args.pg_extras}",
            "AIRFLOW__CELERY__RESULT_BACKEND": f"db+{self.args.pg_driver}://{self.args.pg_user}:{self.args.pg_password}@{self.args.pg_container_name}:{self.args.pg_container_port}/{self.args.pg_schema}{self.args.pg_extras}",
            "AIRFLOW__CELERY__BROKER_URL": f"{self.args.redis_driver}://{self.args.redis_pass}@{self.args.redis_container_name}/{self.args.redis_schema}",
            "AIRFLOW__CORE__FERNET_KEY": "FpErWX7ZxRBGxuAq2JDfle3A7k7Xxi5hY0wh_u0X0Go=",
            "AIRFLOW__CORE__DAGS_ARE_PAUSED_AT_CREATION": "True",
            "AIRFLOW__CORE__LOAD_EXAMPLES": "True",
        }
        if self.args.scheduler_env is not None and isinstance(
            self.args.scheduler_env, dict
        ):
            _scheduler_container_env.update(self.args.scheduler_env)

        _scheduler_container = DockerContainer(
            name=self.args.scheduler_container_name,
            image="{}:{}".format(
                self.args.scheduler_image_name, self.args.scheduler_image_tag
            ),
            command=self.args.scheduler_command,
            # auto_remove=True,
            detach=True,
            entrypoint=self.args.scheduler_entrypoint,
            environment=_scheduler_container_env,
            # remove=True,
            network=docker_build_context.network,
            volumes={
                _airflow_dags_host_dir_path_str: {
                    "bind": self.args.airflow__core__dags_folder,
                    "mode": "rw",
                },
                _airflow_conf_host_dir_path: {
                    "bind": self.args.airflow__core__conf_dir,
                    "mode": "rw",
                },
            },
        )

        _scheduler_rg = DockerResourceGroup(
            name=self.args.scheduler_container_name,
            enabled=self.args.scheduler_enabled,
            weight=102,
            containers=[_scheduler_container],
        )
        # logger.debug("scheduler rg:\n{}".format(_scheduler_rg.json(indent=2)))
        return _scheduler_rg

    def get_webserver_docker_rg(
        self,
        docker_build_context: DockerBuildContext,
    ) -> Optional[DockerResourceGroup]:
        logger.debug(f"Init Webserver DockerResourceGroup")

        _airflow_dags_host_dir_path_str = (
            str(self.args.workspace_dir_path) + self.args.airflow_dags_host_dir_path
        )
        _airflow_conf_host_dir_path = (
            str(self.args.workspace_dir_path) + self.args.airflow_conf_host_dir_path
        )
        _webserver_container_env = {
            "INIT_AIRFLOW_DB": self.args.ws_init_airflow_db,
            "WAIT_FOR_DB": self.args.ws_wait_for_db,
            "WAIT_FOR_REDIS": self.args.ws_wait_for_redis,
            "AIRFLOW_DB_CONN_URL": self.args.pg_container_name,
            "AIRFLOW_DB_CONN_PORT": self.args.pg_container_port,
            "AIRFLOW_DB_USER": self.args.pg_user,
            "AIRFLOW_DB_PASSWORD": self.args.pg_password,
            "AIRFLOW_SCHEMA": self.args.pg_schema,
            "AIRFLOW_REDIS_CONN_URL": self.args.redis_container_name,
            "AIRFLOW_REDIS_CONN_PORT": self.args.redis_container_port,
            "AIRFLOW_REDIS_USER": self.args.redis_user,
            "AIRFLOW_REDIS_PASSWORD": self.args.redis_pass,
            "AIRFLOW_REDIS_SCHEMA": self.args.redis_schema,
            "AIRFLOW__CORE__EXECUTOR": "CeleryExecutor",
            "AIRFLOW__CORE__SQL_ALCHEMY_CONN": f"{self.args.pg_driver}+psycopg2://{self.args.pg_user}:{self.args.pg_password}@{self.args.pg_container_name}:{self.args.pg_container_port}/{self.args.pg_schema}{self.args.pg_extras}",
            "AIRFLOW__CELERY__RESULT_BACKEND": f"db+{self.args.pg_driver}://{self.args.pg_user}:{self.args.pg_password}@{self.args.pg_container_name}:{self.args.pg_container_port}/{self.args.pg_schema}{self.args.pg_extras}",
            "AIRFLOW__CELERY__BROKER_URL": f"{self.args.redis_driver}://{self.args.redis_pass}@{self.args.redis_container_name}/{self.args.redis_schema}",
            "AIRFLOW__CORE__FERNET_KEY": "FpErWX7ZxRBGxuAq2JDfle3A7k7Xxi5hY0wh_u0X0Go=",
            "AIRFLOW__CORE__DAGS_ARE_PAUSED_AT_CREATION": "True",
            "AIRFLOW__CORE__LOAD_EXAMPLES": "True",
        }
        if self.args.ws_env is not None and isinstance(self.args.ws_env, dict):
            _webserver_container_env.update(self.args.ws_env)

        _webserver_container = DockerContainer(
            name=self.args.ws_container_name,
            image="{}:{}".format(self.args.ws_image_name, self.args.ws_image_tag),
            command=self.args.ws_command,
            # auto_remove=True,
            detach=True,
            entrypoint=self.args.ws_entrypoint,
            environment=_webserver_container_env,
            # remove=True,
            network=docker_build_context.network,
            ports={
                self.args.ws_container_port: self.args.ws_container_host_port,
            },
            volumes={
                _airflow_dags_host_dir_path_str: {
                    "bind": self.args.airflow__core__dags_folder,
                    "mode": "rw",
                },
                _airflow_conf_host_dir_path: {
                    "bind": self.args.airflow__core__conf_dir,
                    "mode": "rw",
                },
            },
        )

        _webserver_rg = DockerResourceGroup(
            name=self.args.ws_container_name,
            enabled=self.args.ws_enabled,
            weight=103,
            containers=[_webserver_container],
        )
        # logger.debug("webserver rg:\n{}".format(_webserver_rg.json(indent=2)))
        return _webserver_rg

    def get_airflow_docker_rgs(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[List[DockerResourceGroup]]:
        logger.debug(f"Init Airflow DockerResourceGroups")

        airflow_docker_rgs: List[DockerResourceGroup] = []

        # Airflow Database DockerResourceGroup
        if self.args.db_enabled:
            if self.args.db_type == "postgres":
                _pg_docker_rg = self.get_postgres_docker_rg(docker_build_context)
                if _pg_docker_rg is not None:
                    airflow_docker_rgs.append(_pg_docker_rg)

        # Airflow Redis DockerResourceGroup
        if self.args.redis_enabled:
            _redis_docker_rg = self.get_redis_docker_rg(docker_build_context)
            if _redis_docker_rg is not None:
                airflow_docker_rgs.append(_redis_docker_rg)

        # Airflow Scheduler DockerResourceGroup
        if self.args.scheduler_enabled:
            _scheduler_docker_rg = self.get_scheduler_docker_rg(docker_build_context)
            if _scheduler_docker_rg is not None:
                airflow_docker_rgs.append(_scheduler_docker_rg)

        # Airflow Webserver DockerResourceGroup
        if self.args.ws_enabled:
            _webserver_docker_rg = self.get_webserver_docker_rg(docker_build_context)
            if _webserver_docker_rg is not None:
                airflow_docker_rgs.append(_webserver_docker_rg)

        return airflow_docker_rgs

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        self.docker_resource_groups = OrderedDict()
        # _airflow_rgs: Optional[List[DockerResourceGroup]] = self.get_airflow_docker_rgs(
        #     docker_build_context
        # )
        # if _airflow_rgs is not None:
        #     if self.docker_resource_groups is None:
        #         self.docker_resource_groups = OrderedDict()
        #     for _rg in _airflow_rgs:
        #         self.docker_resource_groups[_rg.name] = _rg



from pathlib import Path
from typing import List, Optional, Dict

from phidata.app import PhidataApp
from phidata.app.databox import default_databox_name
from phidata.infra.aws.resource.eks.cluster import EksCluster
from phidata.infra.base import InfraConfig
from phidata.infra.k8s.args import K8sArgs
from phidata.infra.k8s.manager import K8sManager
from phidata.infra.k8s.resource.group import K8sResourceGroup
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.utils.log import logger


class K8sConfig(InfraConfig):
    def __init__(
        self,
        name: Optional[str] = None,
        env: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
        namespace: str = "default",
        context: Optional[str] = None,
        service_account_name: Optional[str] = None,
        labels: Optional[Dict[str, str]] = None,
        apps: Optional[List[PhidataApp]] = None,
        resources: Optional[List[K8sResourceGroup]] = None,
        create_resources: Optional[List[CreateK8sResourceGroup]] = None,
        resources_dir: str = "k8s",
        databox_name: str = default_databox_name,
        kubeconfig: Optional[str] = None,
        # Get K8s context and kubeconfig from EksCluster
        eks_cluster: Optional[EksCluster] = None,
    ):
        super().__init__()
        try:
            k8s_context = context
            k8s_kubeconfig_path = Path(kubeconfig) if kubeconfig is not None else None
            if eks_cluster is not None:
                # logger.debug("Using EksCluster for K8sConfig")
                k8s_context = eks_cluster.get_kubeconfig_context_name()
                k8s_kubeconfig_path = eks_cluster.kubeconfig_path

            self.args: K8sArgs = K8sArgs(
                name=name,
                env=env,
                version=version,
                enabled=enabled,
                namespace=namespace,
                context=k8s_context,
                service_account_name=service_account_name,
                labels=labels,
                apps=apps,
                resources=resources,
                create_resources=create_resources,
                databox_name=databox_name,
                resources_dir=resources_dir,
                kubeconfig_path=k8s_kubeconfig_path,
                eks_cluster=eks_cluster,
            )
        except Exception as e:
            raise

    @property
    def namespace(self) -> Optional[str]:
        if self.args and self.args.namespace:
            return self.args.namespace
        return None

    @property
    def context(self) -> Optional[str]:
        if self.args and self.args.context:
            return self.args.context
        return None

    @property
    def service_account_name(self) -> Optional[str]:
        if self.args and self.args.service_account_name:
            return self.args.service_account_name
        return None

    @property
    def labels(self) -> Optional[Dict[str, str]]:
        if self.args and self.args.labels:
            return self.args.labels
        return None

    @property
    def apps(self) -> Optional[List[PhidataApp]]:
        if self.args and self.args.apps:
            return self.args.apps
        return None

    @property
    def resources(self) -> Optional[List[K8sResourceGroup]]:
        if self.args and self.args.resources:
            return self.args.resources
        return None

    @property
    def create_resources(self) -> Optional[List[CreateK8sResourceGroup]]:
        if self.args and self.args.create_resources:
            return self.args.create_resources
        return None

    @property
    def resources_dir(self) -> Optional[str]:
        if self.args and self.args.resources_dir:
            return self.args.resources_dir
        return None

    @property
    def databox_name(self) -> Optional[str]:
        if self.args and self.args.databox_name:
            return self.args.databox_name
        return None

    @property
    def kubeconfig_path(self) -> Optional[Path]:
        if self.args and self.args.kubeconfig_path:
            return self.args.kubeconfig_path
        return None

    @property
    def eks_cluster(self) -> Optional[EksCluster]:
        if self.args and self.args.eks_cluster:
            return self.args.eks_cluster
        return None

    def apps_are_valid(self) -> bool:
        if self.apps is None:
            return True
        for _app in self.apps:
            if not isinstance(_app, PhidataApp):
                raise TypeError("Invalid App: {}".format(_app))
        return True

    def resources_are_valid(self) -> bool:
        if self.resources is None:
            return True
        for _resource in self.resources:
            if not isinstance(_resource, K8sResourceGroup):
                raise TypeError("Invalid K8sResourceGroup: {}".format(_resource))
        return True

    def create_resources_are_valid(self) -> bool:
        if self.create_resources is None:
            return True
        for _resource in self.create_resources:
            if not isinstance(_resource, CreateK8sResourceGroup):
                raise TypeError("Invalid CreateK8sResourceGroup: {}".format(_resource))
        return True

    def is_valid(self) -> bool:
        return (
            self.apps_are_valid()
            and self.resources_are_valid()
            and self.create_resources_are_valid()
        )

    def get_k8s_manager(self) -> Optional[K8sManager]:
        return K8sManager(k8s_args=self.args)

    def get_app_by_name(self, app_name: str) -> Optional[PhidataApp]:

        if self.apps is None:
            return None

        for _app in self.apps:
            try:
                if app_name == _app.name:
                    return _app
            except Exception:
                continue
        return None



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, Any, List, Union

from pydantic import BaseModel

from phidata.app import PhidataApp, PhidataAppArgs
from phidata.infra.docker.resource.network import DockerNetwork
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.k8s.create.apps.v1.deployment import CreateDeployment
from phidata.infra.k8s.create.core.v1.config_map import CreateConfigMap
from phidata.infra.k8s.create.core.v1.container import CreateContainer
from phidata.infra.k8s.create.core.v1.volume import (
    CreateVolume,
    EmptyDirVolumeSource,
    HostPathVolumeSource,
    VolumeType,
)
from phidata.infra.k8s.create.common.port import CreatePort
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.common import get_default_configmap_name
from phidata.utils.cli_console import print_error, print_info
from phidata.utils.log import logger

default_devbox_name: str = "devbox"


class DevboxDevModeArgs(BaseModel):
    install_phidata_dev: bool = True
    phidata_volume_name: str = "devbox-phidata-volume"
    phidata_dir_path: Path = Path.home().joinpath("p", "philab", "phidata")
    phidata_dir_container_path: str = "/phidata"


class DevboxArgs(PhidataAppArgs):
    name: str = default_devbox_name
    version: str = "1"
    enabled: bool = True

    # Image args
    image_name: str = "phidata/devbox"
    image_tag: str = "1"
    # Path to entrypoint script in the image
    entrypoint: Optional[Union[str, List]] = None
    command: Optional[Union[str, List]] = None
    # Add env variables to container env
    # key: var name, value: var value
    env: Optional[Dict[str, str]] = None

    ## Configure python packages
    # Install python dependencies using a requirements.txt file
    install_requirements: bool = True
    requirements_volume_name: str = "devbox-requirements-volume"
    # requirements.txt file path from the workspace root (on the host)
    requirements_file_path: str = "workspace/requirements/dev-requirements.txt"
    # requirements.txt file path in container
    requirements_file_container_path: str = "/requirements.txt"

    ## Configure the Data volume
    # Create a data directory on the container
    create_data_volume: bool = True
    data_dir_volume_name: str = "devbox-data-volume"
    # Container path to mount the data volume
    data_dir_container_path: str = "/usr/local/data"
    # If True, create the data volume using an empty dir
    #   ignored if use_host_path_for_data_volume = True
    use_empty_dir_for_data_volume: bool = True
    # If True, use a host path for the data volume
    use_host_path_for_data_volume: bool = False
    # Path for data dir from the workspace root (on the host)
    data_dir_path: Optional[str] = "data"

    ## Configure the Workspace volume
    # Mount the workspace directory on the container
    mount_workspace: bool = True
    workspace_volume_name: str = "devbox-workspace-volume"
    # Container path to mount the workspace volume in
    # This path is the parent directory in which the workspace directory will be mounted
    # eg: if ws name is: idata, full workspace path would be: /usr/local/idata
    workspace_parent_container_path: str = "/usr/local"
    # Path to projects dir in the workspace directory
    # Used to join with workspace_parent_container_path to derive PHI_PROJECTS_DIR variable
    projects_dir: str = "projects"
    # If True, use a host path for the workspace volume
    use_host_path_for_workspace: bool = True

    ## Configure Aws credentials
    # Mount Aws credentials
    mount_aws_credentials: bool = False
    # credentials file path on the host
    aws_credentials_dir_path: Optional[str] = None
    # credentials file path on the container
    aws_credentials_dir_container_path: str = "/usr/local/devbox/.aws"

    ## Configure the Airflow volume
    # True => mount the airflow_dir_path to airflow_dir_container_file_path
    mount_airflow_dir: bool = False
    airflow_volume_name: str = "devbox-airflow-volume"
    # Container path to mount the airflow volume
    airflow_dir_container_path: str = "/usr/local/airflow"
    # Airflow dir path from the workspace root (on the host)
    airflow_dir_path: str = "workspace/airflow"

    ## Airflow args
    # If init_airflow = True, this devbox sets up airflow
    init_airflow: bool = True
    # If use_projects_for_airflow_dags = True,
    # set the AIRFLOW__CORE__DAGS_FOLDER to the projects_dir_container_path
    use_projects_for_airflow_dags: bool = True
    # airflow_dags_dir_path: The directory in the container containing the airflow pipelines
    # If use_projects_for_airflow_dags = False,
    # set the AIRFLOW__CORE__DAGS_FOLDER to the airflow_dags_dir_path
    airflow_dags_dir_path: Optional[str] = None

    # Airflow db args
    # If True, initialized the airflow_db on this devbox
    # If None, value is derived from init_airflow i.e. initialize the airflow_db if init_airflow = True
    #   init_airflow_db = None is used to init airflow_db on sqllite
    # If using the devbox with an existing Airflow db, set init_airflow_db = False
    init_airflow_db: Optional[bool] = None
    wait_for_airflow_db: bool = False
    airflow_db_conn_url: Optional[str] = "airflow-pg"
    airflow_db_conn_port: Optional[str] = "5432"
    airflow_db_user: Optional[str] = "airflow"
    airflow_db_password: Optional[str] = "airflow"
    airflow_schema: Optional[str] = "airflow"
    # Airflow db connections in the format { conn_id: conn_url }
    # converted to env var: AIRFLOW_CONN__conn_id = conn_url
    db_connections: Optional[Dict] = None

    # Airflow scheduler args
    # Init Airflow scheduler as a deamon process
    init_airflow_scheduler: bool = False

    # Airflow webserver args
    # Init Airflow webserver when the container starts
    init_airflow_webserver: bool = False
    # webserver port on the container
    airflow_webserver_container_port: int = 8080
    # webserver port on the host machine
    airflow_webserver_host_port: int = 8080
    ## K8s args
    # webserver port name
    airflow_webserver_port_name: str = "http"

    # Other airflow args
    load_examples: bool = False

    # Other args
    print_env_on_load: bool = True
    dev_mode: Optional[DevboxDevModeArgs] = None


class Devbox(PhidataApp):
    def __init__(
        self,
        name: str = default_devbox_name,
        version: str = "1",
        enabled: bool = True,
        # Image args
        image_name: str = "phidata/devbox",
        image_tag: str = "1",
        # Path to entrypoint script in the image
        entrypoint: Optional[Union[str, List]] = None,
        command: Optional[Union[str, List]] = None,
        # Add env variables to container env
        # key: var name, value: var value
        env: Optional[Dict[str, str]] = None,
        ## Configure python packages to install
        # Install python dependencies using a requirements.txt file
        install_requirements: bool = True,
        requirements_volume_name: str = "devbox-requirements-volume",
        # requirements.txt file path from the workspace root (on the host)
        requirements_file_path: str = "workspace/requirements/dev-requirements.txt",
        # requirements.txt file path in container
        requirements_file_container_path: str = "/requirements.txt",
        ## Configure the Data volume,
        # Create a data directory on the container
        create_data_volume: bool = True,
        data_volume_name: str = "devbox-data-volume",
        # Container path to mount the data volume
        data_dir_container_path: str = "/usr/local/data",
        # If True, create the data volume using an empty dir
        #   ignored if use_host_path_for_data_volume = True
        use_empty_dir_for_data_volume: bool = True,
        # If True, use a host path for the data volume
        use_host_path_for_data_volume: bool = False,
        # Path for data dir from the workspace root (on the host)
        data_dir_path: Optional[str] = "data",
        ## Configure the Workspace volume
        # Create a workspace directory on the container
        mount_workspace: bool = True,
        workspace_volume_name: str = "devbox-workspace-volume",
        # Container path to mount the workspace volume
        workspace_parent_container_path: str = "/usr/local/",
        # Path to projects dir in the workspace directory
        # Used to join with workspace_parent_container_path to derive PHI_PROJECTS_DIR variable
        projects_dir: str = "projects",
        # If True, use a host path for the workspace volume
        use_host_path_for_workspace: bool = True,
        ## Configure Aws credentials
        # Mount Aws credentials
        mount_aws_credentials: bool = False,
        # credentials file path on the host
        aws_credentials_dir_path: Optional[str] = None,
        # credentials file path on the container
        aws_credentials_dir_container_path: str = "/usr/local/devbox/.aws",
        ## Configure the Airflow volume
        # True => mount the airflow_dir_path to airflow_dir_container_file_path
        mount_airflow_dir: bool = False,
        airflow_volume_name: str = "devbox-airflow-volume",
        # Container path to mount the airflow volume
        airflow_dir_container_path: str = "/usr/local/airflow",
        # Airflow dir path from the workspace root (on the host)
        airflow_dir_path: str = "workspace/airflow",
        ## Airflow args,
        # If init_airflow = True, this devbox sets up airflow,
        init_airflow: bool = True,
        # If use_projects_for_airflow_dags = True,,
        # set the AIRFLOW__CORE__DAGS_FOLDER to the projects_dir_container_path,
        use_projects_for_airflow_dags: bool = True,
        # airflow_dags_dir_path: The directory in the container containing the airflow pipelines,
        # If use_projects_for_airflow_dags = False,,
        # set the AIRFLOW__CORE__DAGS_FOLDER to the airflow_dags_dir_path,
        airflow_dags_dir_path: Optional[str] = None,
        # Airflow db args,
        # If True, initialized the airflow_db on this devbox,
        # If None, value is derived from init_airflow i.e. initialize the airflow_db if init_airflow = True,
        #   init_airflow_db = None is used to init airflow_db ~ sqllite,
        # If using the devbox with an existing Airflow db, set init_airflow_db = False,
        init_airflow_db: Optional[bool] = None,
        wait_for_airflow_db: bool = False,
        airflow_db_conn_url: Optional[str] = "airflow-pg",
        airflow_db_conn_port: Optional[str] = "5432",
        airflow_db_user: Optional[str] = "airflow",
        airflow_db_password: Optional[str] = "airflow",
        airflow_schema: Optional[str] = "airflow",
        # Airflow db connections in the format { conn_id: conn_url },
        # converted to env var: AIRFLOW_CONN__conn_id = conn_url,
        db_connections: Optional[Dict] = None,
        # Airflow scheduler args,
        # Init Airflow scheduler as a deamon process,
        init_airflow_scheduler: bool = False,
        # Airflow webserver args,
        # Init Airflow webserver when the container starts
        init_airflow_webserver: bool = False,
        # webserver port on the container,
        airflow_webserver_container_port: int = 8080,
        # webserver port on the host machine,
        airflow_webserver_host_port: int = 8080,
        ## K8s args,
        # webserver port name,
        airflow_webserver_port_name: str = "http",
        # Other airflow args,
        load_examples: bool = False,
        # Other args,
        print_env_on_load: bool = True,
        dev_mode: Optional[DevboxDevModeArgs] = None,
        # Additional args
        # If True, skip resource creation if active resources with the same name exist.
        use_cache: bool = True,
        # If True, log extra debug messages
        use_verbose_logs: bool = False,
    ):

        super().__init__()
        try:
            self.args: DevboxArgs = DevboxArgs(
                name=name,
                version=version,
                enabled=enabled,
                image_name=image_name,
                image_tag=image_tag,
                entrypoint=entrypoint,
                command=command,
                env=env,
                install_requirements=install_requirements,
                requirements_volume_name=requirements_volume_name,
                requirements_file_path=requirements_file_path,
                requirements_file_container_path=requirements_file_container_path,
                create_data_volume=create_data_volume,
                data_dir_volume_name=data_volume_name,
                data_dir_container_path=data_dir_container_path,
                use_empty_dir_for_data_volume=use_empty_dir_for_data_volume,
                use_host_path_for_data_volume=use_host_path_for_data_volume,
                data_dir_path=data_dir_path,
                mount_workspace=mount_workspace,
                workspace_volume_name=workspace_volume_name,
                workspace_parent_container_path=workspace_parent_container_path,
                projects_dir=projects_dir,
                use_host_path_for_workspace=use_host_path_for_workspace,
                mount_aws_credentials=mount_aws_credentials,
                aws_credentials_dir_path=aws_credentials_dir_path,
                aws_credentials_dir_container_path=aws_credentials_dir_container_path,
                mount_airflow_dir=mount_airflow_dir,
                airflow_volume_name=airflow_volume_name,
                airflow_dir_container_path=airflow_dir_container_path,
                airflow_dir_path=airflow_dir_path,
                init_airflow=init_airflow,
                use_projects_for_airflow_dags=use_projects_for_airflow_dags,
                airflow_dags_dir_path=airflow_dags_dir_path,
                init_airflow_db=init_airflow_db,
                wait_for_airflow_db=wait_for_airflow_db,
                airflow_db_conn_url=airflow_db_conn_url,
                airflow_db_conn_port=airflow_db_conn_port,
                airflow_db_user=airflow_db_user,
                airflow_db_password=airflow_db_password,
                airflow_schema=airflow_schema,
                db_connections=db_connections,
                init_airflow_scheduler=init_airflow_scheduler,
                init_airflow_webserver=init_airflow_webserver,
                airflow_webserver_container_port=airflow_webserver_container_port,
                airflow_webserver_host_port=airflow_webserver_host_port,
                airflow_webserver_port_name=airflow_webserver_port_name,
                load_examples=load_examples,
                print_env_on_load=print_env_on_load,
                dev_mode=dev_mode,
                use_cache=use_cache,
                use_verbose_logs=use_verbose_logs,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    ######################################################
    ## Docker Resources
    ######################################################

    def init_airflow_on_docker_container(self, container: DockerContainer) -> None:
        """
        Initialize airflow on a docker container
        """

        if not self.args.init_airflow:
            return

        # Update init_airflow_db arg if None
        if self.args.init_airflow_db is None:
            self.args.init_airflow_db = True

        airflow_env: Dict[str, Any] = {
            "INIT_AIRFLOW": str(self.args.init_airflow),
            "INIT_AIRFLOW_DB": str(self.args.init_airflow_db),
            "WAIT_FOR_AIRFLOW_DB": str(self.args.wait_for_airflow_db),
            "AIRFLOW_DB_CONN_URL": self.args.airflow_db_conn_url,
            "AIRFLOW_DB_CONN_PORT": self.args.airflow_db_conn_port,
            "AIRFLOW_DB_USER": self.args.airflow_db_user,
            "AIRFLOW_DB_PASSWORD": self.args.airflow_db_password,
            "AIRFLOW_SCHEMA": self.args.airflow_schema,
            "INIT_AIRFLOW_SCHEDULER": str(self.args.init_airflow_scheduler),
            "INIT_AIRFLOW_WEBSERVER": str(self.args.init_airflow_webserver),
            "CREATE_AIRFLOW_TEST_USER": str(self.args.init_airflow_webserver),
            "AIRFLOW__CORE__LOAD_EXAMPLES": str(self.args.load_examples),
        }
        # Set the AIRFLOW__CORE__DAGS_FOLDER
        if self.args.mount_workspace and self.args.use_projects_for_airflow_dags:
            projects_dir_container_path = Path(
                self.args.workspace_parent_container_path
            ).joinpath(self.args.workspace_dir_path.stem, self.args.projects_dir)
            airflow_env["AIRFLOW__CORE__DAGS_FOLDER"] = projects_dir_container_path
        elif self.args.airflow_dags_dir_path is not None:
            airflow_env["AIRFLOW__CORE__DAGS_FOLDER"] = self.args.airflow_dags_dir_path

        # Set the AIRFLOW__CONN_ variables
        if self.args.db_connections is not None:
            for conn_id, conn_url in self.args.db_connections.items():
                try:
                    af_conn_id = str("AIRFLOW_CONN_{}".format(conn_id)).upper()
                    airflow_env[af_conn_id] = conn_url
                except Exception as e:
                    logger.exception(e)
                    continue
        # Update the container.environment to include airflow_env
        container.environment.update(airflow_env)

        # Open the airflow webserver port if init_airflow_webserver = True
        if self.args.init_airflow_webserver:
            ws_port: Dict[str, Any] = {
                str(
                    self.args.airflow_webserver_container_port
                ): self.args.airflow_webserver_host_port,
            }
            if container.ports is None:
                container.ports = {}
            container.ports.update(ws_port)

    def get_devbox_docker_rg(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[DockerResourceGroup]:
        logger.debug(f"Init Devbox DockerResourceGroup")

        # Set the Devbox Env
        workspace_dir_container_path = Path(
            self.args.workspace_parent_container_path
        ).joinpath(self.args.workspace_dir_path.stem)
        projects_dir_container_path = workspace_dir_container_path.joinpath(
            self.args.projects_dir
        )
        devbox_container_env: Dict[str, Any] = {
            # Env variables used by data workflows and data assets
            "PHI_DATA_DIR": self.args.data_dir_container_path,
            "PHI_PROJECTS_DIR": str(projects_dir_container_path),
            "PHI_WORKSPACE_DIR": str(workspace_dir_container_path),
            "PHI_WORKSPACE_PARENT": str(self.args.workspace_parent_container_path),
            # Configure python packages to install
            "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
            "REQUIREMENTS_FILE_PATH": self.args.requirements_file_container_path,
            # Print these env variables when the container is loaded
            "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
            # Dev mode args
            "INSTALL_PHIDATA_DEV": str(self.args.dev_mode.install_phidata_dev)
            if self.args.dev_mode is not None
            else str(False),
            "PHIDATA_DIR_PATH": self.args.dev_mode.phidata_dir_container_path
            if self.args.dev_mode is not None
            else str(False),
        }
        # logger.debug(f"devbox_container_env: {devbox_container_env}")
        # Update the devbox env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            devbox_container_env.update(self.args.env)

        # Devbox Volumes
        # devbox_volumes: A dictionary to configure volumes mounted inside the container.
        # The key is either the host path or a volume name,
        # and the value is a dictionary with 2 keys:
        #   bind - The path to mount the volume inside the container
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
        # For example:
        # {
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
        # }
        devbox_volumes = {}
        # Create a volume for the requirements file
        # if install_requirements = True
        if self.args.install_requirements:
            # Use requirements_file_absolute_path as the dict key
            requirements_file_absolute_path = str(
                self.args.workspace_dir_path.joinpath(self.args.requirements_file_path)
            )
            logger.debug(f"Mounting: {requirements_file_absolute_path}")
            logger.debug(f"\tto: {self.args.requirements_file_container_path}")
            devbox_volumes[requirements_file_absolute_path] = {
                "bind": self.args.requirements_file_container_path,
                "mode": "rw",
            }
        # Create a volume for the data dir
        # if create_data_volume = True
        if self.args.create_data_volume:
            data_dir_volume_key: Optional[str] = None
            if (
                self.args.use_host_path_for_data_volume
                and self.args.data_dir_path is not None
            ):
                # Use data_dir_absolute_path as the dict key
                data_dir_volume_key = str(
                    self.args.workspace_dir_path.joinpath(self.args.data_dir_path)
                )
            elif self.args.use_empty_dir_for_data_volume:
                # Use volume name as the dict key
                data_dir_volume_key = self.args.data_dir_volume_name
            if data_dir_volume_key is not None:
                logger.debug(f"Mounting: {data_dir_volume_key}")
                logger.debug(f"\tto: {self.args.data_dir_container_path}")
                devbox_volumes[data_dir_volume_key] = {
                    "bind": self.args.data_dir_container_path,
                    "mode": "rw",
                }
        # Create a volume for the workspace dir
        # if mount_workspace = True
        if self.args.mount_workspace:
            if self.args.use_host_path_for_workspace:
                ws_dir_container_path_str = str(
                    Path(self.args.workspace_parent_container_path).joinpath(
                        self.args.workspace_dir_path.stem
                    )
                )
                ws_dir_path_str = str(self.args.workspace_dir_path)
                logger.debug(f"Mounting: {ws_dir_path_str}")
                logger.debug(f"\tto: {ws_dir_container_path_str}")
                devbox_volumes[ws_dir_path_str] = {
                    "bind": ws_dir_container_path_str,
                    "mode": "rw",
                }
        # Create a volume for Aws credentials
        if self.args.mount_aws_credentials:
            # Use requirements_file_absolute_path as the dict key
            aws_credentials_dir_path = (
                self.args.aws_credentials_dir_path
                if self.args.aws_credentials_dir_path is not None
                else str(Path.home().resolve().joinpath(".aws"))
            )
            logger.debug(f"Mounting: {aws_credentials_dir_path}")
            logger.debug(f"\tto: {self.args.aws_credentials_dir_container_path}")
            devbox_volumes[aws_credentials_dir_path] = {
                "bind": self.args.aws_credentials_dir_container_path,
                "mode": "ro",
            }
            devbox_container_env[
                "AWS_CONFIG_FILE"
            ] = f"{self.args.aws_credentials_dir_container_path}/config"
            devbox_container_env[
                "AWS_SHARED_CREDENTIALS_FILE"
            ] = f"{self.args.aws_credentials_dir_container_path}/credentials"
        # Create a volume for the Airflow conf file
        # if mount_airflow_dir = True
        if self.args.mount_airflow_dir:
            airflow_dir_file_absolute_path = str(
                self.args.workspace_dir_path.joinpath(self.args.airflow_dir_path)
            )
            logger.debug(f"Mounting: {airflow_dir_file_absolute_path}")
            logger.debug(f"\tto: {self.args.airflow_dir_container_path}")
            devbox_volumes[airflow_dir_file_absolute_path] = {
                "bind": self.args.airflow_dir_container_path,
                "mode": "rw",
            }
        # Create a volume for Phidata dev mode
        # if dev_mode is enabled
        if self.args.dev_mode is not None:
            # mount the phidata directory if install_phidata_dev = True
            if (
                self.args.dev_mode.install_phidata_dev
                and self.args.dev_mode.phidata_dir_path is not None
            ):
                phidata_dir_absolute_path = str(self.args.dev_mode.phidata_dir_path)
                logger.debug(f"Mounting: {phidata_dir_absolute_path}")
                logger.debug(f"\tto: {self.args.dev_mode.phidata_dir_container_path}")
                devbox_volumes[phidata_dir_absolute_path] = {
                    "bind": self.args.dev_mode.phidata_dir_container_path,
                    "mode": "rw",
                }

        # Create the devbox container
        devbox_docker_container = DockerContainer(
            name=self.args.name,
            image=f"{self.args.image_name}:{self.args.image_tag}",
            command=self.args.command,
            auto_remove=True,
            detach=True,
            entrypoint=self.args.entrypoint,
            environment=devbox_container_env,
            remove=True,
            stdin_open=True,
            tty=True,
            network=docker_build_context.network,
            volumes=devbox_volumes,
            use_cache=self.args.use_cache,
            use_verbose_logs=self.args.use_verbose_logs,
        )
        # Initialize airflow on container
        self.init_airflow_on_docker_container(devbox_docker_container)

        devbox_docker_resource_group = DockerResourceGroup(
            name=self.args.name,
            enabled=self.args.enabled,
            network=DockerNetwork(name=docker_build_context.network),
            containers=[devbox_docker_container],
        )
        return devbox_docker_resource_group

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        devbox_docker_rg: Optional[DockerResourceGroup] = self.get_devbox_docker_rg(
            docker_build_context
        )
        if devbox_docker_rg is not None:
            if self.docker_resource_groups is None:
                self.docker_resource_groups = OrderedDict()
            self.docker_resource_groups[devbox_docker_rg.name] = devbox_docker_rg

    ######################################################
    ## K8s Resources
    ######################################################

    def init_airflow_on_k8s_container(
        self, container: CreateContainer, k8s_resource_group: CreateK8sResourceGroup
    ) -> None:
        """
        Initialize airflow on a k8s container
        """

        if not self.args.init_airflow:
            return

        # Update init_airflow_db arg if None
        if self.args.init_airflow_db is None:
            self.args.init_airflow_db = True

        airflow_cm = CreateConfigMap(
            cm_name=get_default_configmap_name("devbox-airflow"),
            app_name=self.args.name,
            data={
                "INIT_AIRFLOW": str(self.args.init_airflow),
                "INIT_AIRFLOW_DB": str(self.args.init_airflow_db),
                "WAIT_FOR_AIRFLOW_DB": str(self.args.wait_for_airflow_db),
                "AIRFLOW_DB_CONN_URL": self.args.airflow_db_conn_url,
                "AIRFLOW_DB_CONN_PORT": self.args.airflow_db_conn_port,
                "AIRFLOW_DB_USER": self.args.airflow_db_user,
                "AIRFLOW_DB_PASSWORD": self.args.airflow_db_password,
                "AIRFLOW_SCHEMA": self.args.airflow_schema,
                "INIT_AIRFLOW_SCHEDULER": str(self.args.init_airflow_scheduler),
                "INIT_AIRFLOW_WEBSERVER": str(self.args.init_airflow_webserver),
                "CREATE_AIRFLOW_TEST_USER": str(self.args.init_airflow_webserver),
                "AIRFLOW__CORE__LOAD_EXAMPLES": str(self.args.load_examples),
            },
        )

        # Add airflow_cm to container env and
        if container.envs_from_configmap is None:
            container.envs_from_configmap = []
        container.envs_from_configmap.append(airflow_cm.cm_name)
        if k8s_resource_group.config_maps is None:
            k8s_resource_group.config_maps = []
        k8s_resource_group.config_maps.append(airflow_cm)

        # Set the AIRFLOW__CORE__DAGS_FOLDER
        if self.args.mount_workspace and self.args.use_projects_for_airflow_dags:
            projects_dir_container_path = Path(
                self.args.workspace_parent_container_path
            ).joinpath(self.args.workspace_dir_path.stem, self.args.projects_dir)
            airflow_cm.data["AIRFLOW__CORE__DAGS_FOLDER"] = projects_dir_container_path
        elif self.args.airflow_dags_dir_path is not None:
            airflow_cm.data[
                "AIRFLOW__CORE__DAGS_FOLDER"
            ] = self.args.airflow_dags_dir_path
        if self.args.db_connections is not None:
            for conn_id, conn_url in self.args.db_connections.items():
                try:
                    af_conn_id = str("AIRFLOW_CONN_{}".format(conn_id)).upper()
                    airflow_cm.data[af_conn_id] = conn_url
                except Exception as e:
                    logger.exception(e)
                    continue
        # Open the airflow webserver port if init_airflow_webserver = True
        if self.args.init_airflow_webserver:
            ws_port = CreatePort(
                name=self.args.airflow_webserver_port_name,
                container_port=self.args.airflow_webserver_container_port,
            )
            if container.ports is None:
                container.ports = []
            container.ports.append(ws_port)

    def get_devbox_k8s_rg(
        self, k8s_build_context: K8sBuildContext
    ) -> Optional[K8sResourceGroup]:
        logger.debug(f"Init Devbox K8sResourceGroup")

        # Set the Devbox Env
        # Devbox Config-Map is a dict of environment variables which are not secrets.
        # The keys of this configmap will become env variables in the devbox container
        workspace_dir_container_path = Path(
            self.args.workspace_parent_container_path
        ).joinpath(self.args.workspace_dir_path.stem)
        projects_dir_container_path = workspace_dir_container_path.joinpath(
            self.args.projects_dir
        )
        devbox_cm = CreateConfigMap(
            cm_name=get_default_configmap_name(self.args.name),
            app_name=self.args.name,
            data={
                # Env variables used by data workflows and data assets
                "PHI_DATA_DIR": self.args.data_dir_container_path,
                "PHI_PROJECTS_DIR": str(projects_dir_container_path),
                "PHI_WORKSPACE_DIR": str(workspace_dir_container_path),
                "PHI_WORKSPACE_PARENT": str(self.args.workspace_parent_container_path),
                # Configure python packages to install
                "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
                "REQUIREMENTS_FILE_PATH": self.args.requirements_file_container_path,
                # Print these env variables when the container is loaded
                "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
                # Dev mode args
                "INSTALL_PHIDATA_DEV": str(self.args.dev_mode.install_phidata_dev)
                if self.args.dev_mode is not None
                else str(False),
                "PHIDATA_DIR_PATH": self.args.dev_mode.phidata_dir_container_path
                if self.args.dev_mode is not None
                else str(False),
            },
        )
        # Update the devbox env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            devbox_cm.data.update(self.args.env)

        # Devbox Volumes
        devbox_volumes = []
        # Create a volume for the requirements file
        # if install_requirements = True
        if self.args.install_requirements:
            requirements_file_absolute_path = str(
                self.args.workspace_dir_path.joinpath(self.args.requirements_file_path)
            )
            logger.debug(f"Mounting: {requirements_file_absolute_path}")
            logger.debug(f"\tto: {self.args.requirements_file_container_path}")
            requirements_volume = CreateVolume(
                volume_name=self.args.requirements_volume_name,
                app_name=self.args.name,
                mount_path=self.args.requirements_file_container_path,
                volume_type=VolumeType.HOST_PATH,
                host_path=HostPathVolumeSource(path=requirements_file_absolute_path),
            )
            devbox_volumes.append(requirements_volume)
        # Create a volume for the data dir
        # if create_data_volume = True
        if self.args.create_data_volume:
            if (
                self.args.use_host_path_for_data_volume
                and self.args.data_dir_path is not None
            ):
                data_dir_absolute_path = str(
                    self.args.workspace_dir_path.joinpath(self.args.data_dir_path)
                )
                logger.debug(f"Mounting: {data_dir_absolute_path}")
                logger.debug(f"\tto: {self.args.data_dir_container_path}")
                data_volume = CreateVolume(
                    volume_name=self.args.data_dir_volume_name,
                    app_name=self.args.name,
                    mount_path=self.args.data_dir_container_path,
                    volume_type=VolumeType.HOST_PATH,
                    host_path=HostPathVolumeSource(
                        path=data_dir_absolute_path,
                    ),
                )
                devbox_volumes.append(data_volume)
            elif self.args.use_empty_dir_for_data_volume:
                logger.debug(f"Mounting EmptyDir")
                logger.debug(f"\tto: {self.args.data_dir_container_path}")
                data_volume = CreateVolume(
                    volume_name=self.args.data_dir_volume_name,
                    app_name=self.args.name,
                    mount_path=self.args.data_dir_container_path,
                    volume_type=VolumeType.EMPTY_DIR,
                    empty_dir=EmptyDirVolumeSource(),
                )
                devbox_volumes.append(data_volume)
        # Create a volume for the projects dir
        # if mount_workspace = True
        if self.args.mount_workspace:
            if self.args.use_host_path_for_workspace:
                ws_dir_container_path_str = str(
                    Path(self.args.workspace_parent_container_path).joinpath(
                        self.args.workspace_dir_path.stem
                    )
                )
                ws_dir_path_str = str(self.args.workspace_dir_path)
                logger.debug(f"Mounting: {ws_dir_path_str}")
                logger.debug(f"\tto: {ws_dir_container_path_str}")
                workspace_volume = CreateVolume(
                    volume_name=self.args.workspace_volume_name,
                    app_name=self.args.name,
                    mount_path=ws_dir_container_path_str,
                    volume_type=VolumeType.HOST_PATH,
                    host_path=HostPathVolumeSource(
                        path=ws_dir_path_str,
                    ),
                )
                devbox_volumes.append(workspace_volume)
        # Airflow conf file
        if self.args.mount_airflow_dir:
            airflow_dir_file_absolute_path = str(
                self.args.workspace_dir_path.joinpath(self.args.airflow_dir_path)
            )
            logger.debug(f"Mounting: {airflow_dir_file_absolute_path}")
            logger.debug(f"\tto: {self.args.airflow_dir_container_path}")
            airflow_dir_volume = CreateVolume(
                volume_name=self.args.airflow_volume_name,
                app_name=self.args.name,
                mount_path=self.args.airflow_dir_container_path,
                volume_type=VolumeType.HOST_PATH,
                host_path=HostPathVolumeSource(path=airflow_dir_file_absolute_path),
            )
            devbox_volumes.append(airflow_dir_volume)
        # Create a volume for Phidata dev mode
        # if dev_mode is enabled
        if self.args.dev_mode is not None:
            # mount the phidata directory if install_phidata_dev = True
            if (
                self.args.dev_mode.install_phidata_dev
                and self.args.dev_mode.phidata_dir_path is not None
            ):
                phidata_dir_absolute_path = str(self.args.dev_mode.phidata_dir_path)
                logger.debug(f"Mounting: {phidata_dir_absolute_path}")
                logger.debug(f"\tto: {self.args.dev_mode.phidata_dir_container_path}")
                phidata_volume = CreateVolume(
                    volume_name=self.args.dev_mode.phidata_volume_name,
                    app_name=self.args.name,
                    mount_path=self.args.dev_mode.phidata_dir_container_path,
                    volume_type=VolumeType.HOST_PATH,
                    host_path=HostPathVolumeSource(
                        path=phidata_dir_absolute_path,
                    ),
                )
                devbox_volumes.append(phidata_volume)

        # Create the devbox container
        devbox_k8s_container = CreateContainer(
            container_name=f"{self.args.name}-container",
            app_name=self.args.name,
            image_name=self.args.image_name,
            image_tag=self.args.image_tag,
            # Equivalent to docker images CMD
            args=self.args.command,
            # Equivalent to docker images ENTRYPOINT
            command=self.args.entrypoint,
            envs_from_configmap=[devbox_cm.cm_name],
            volumes=devbox_volumes,
        )
        devbox_deployment = CreateDeployment(
            deploy_name=f"{self.args.name}-deploy",
            pod_name=f"{self.args.name}-pod",
            app_name=self.args.name,
            namespace=k8s_build_context.namespace,
            service_account_name=k8s_build_context.service_account_name,
            containers=[devbox_k8s_container],
            volumes=devbox_volumes,
            labels=k8s_build_context.labels,
        )
        devbox_k8s_resource_group = CreateK8sResourceGroup(
            name=self.args.name,
            enabled=self.args.enabled,
            config_maps=[devbox_cm],
            deployments=[devbox_deployment],
        )
        # Initialize airflow on container
        self.init_airflow_on_k8s_container(
            devbox_k8s_container, devbox_k8s_resource_group
        )
        return devbox_k8s_resource_group.create()

    def init_k8s_resource_groups(self, k8s_build_context: K8sBuildContext) -> None:
        devbox_k8s_rg: Optional[K8sResourceGroup] = self.get_devbox_k8s_rg(
            k8s_build_context
        )
        # logger.debug("devbox_k8s_rg: {}".format(devbox_k8s_rg))
        if devbox_k8s_rg is not None:
            if self.k8s_resource_groups is None:
                self.k8s_resource_groups = OrderedDict()
            self.k8s_resource_groups[devbox_k8s_rg.name] = devbox_k8s_rg



from pathlib import Path
from typing import Optional, Union, Any

from phidata.asset import DataAsset, DataAssetArgs
from phidata.utils.cli_console import print_error
from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger
from phidata.utils.filesystem import delete_from_fs


class FileType(ExtendedEnum):
    CSV = "CSV"
    TSV = "TSV"
    TXT = "TXT"
    JSON = "JSON"


class FileArgs(DataAssetArgs):
    name: Optional[str]
    # parent directory of the file relative to the base_dir
    file_dir: Optional[Union[str, Path]]
    file_type: Optional[FileType]
    # absolute path for the file
    file_path: Optional[Path]


class File(DataAsset):
    def __init__(
        self,
        name: Optional[str] = None,
        # parent directory of the file relative to the base_dir
        file_dir: Optional[Union[str, Path]] = None,
        file_type: Optional[FileType] = None,
        # absolute path for the file
        file_path: Optional[Path] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ) -> None:

        super().__init__()
        try:
            self.args: FileArgs = FileArgs(
                name=name,
                file_dir=file_dir,
                file_type=file_type,
                file_path=file_path,
                version=version,
                enabled=enabled,
            )
        except Exception as e:
            print_error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def file_dir(self) -> Optional[Union[str, Path]]:
        return self.args.file_dir

    @file_dir.setter
    def file_dir(self, file_dir: Union[str, Path]) -> None:
        if file_dir is not None:
            self.args.file_dir = file_dir

    @property
    def file_type(self) -> Optional[FileType]:
        if self.args.file_type is not None:
            return self.args.file_type

        # logger.debug("-*--*- Building file_type")
        fp = self.file_path
        if fp is None:
            return None
        suffix = fp.suffix[1:]  # remove the . from ".json" or ".csv"
        # logger.debug("suffix: {}".format(suffix))

        derived_file_type: Optional[FileType] = None
        if suffix.upper() in FileType.values_list():
            derived_file_type = FileType.from_str(suffix.upper())
            self.args.file_type = derived_file_type
        return self.args.file_type

    @file_type.setter
    def file_type(self, file_type: Optional[FileType]) -> None:
        if file_type is not None:
            self.args.file_type = file_type

    @property
    def file_path(self) -> Optional[Path]:
        if self.args.file_path is not None:
            return self.args.file_path

        # logger.debug("-*--*- Building file_path")
        # Start with the base_dir_path
        # base_dir_path is loaded from the current environment variable
        _file_path: Optional[Path] = self.base_dir_path
        # Add the file_dir if available
        if self.file_dir is not None:
            if _file_path is None:
                # _file_path is None meaning no base_dir_path
                _file_path = Path(self.file_dir)
            else:
                _file_path = _file_path.joinpath(self.file_dir)

        # Add the file_name
        if self.name is not None:
            if _file_path is None:
                _file_path = Path(self.name)
            else:
                _file_path = _file_path.joinpath(self.name)

        self.args.file_path = _file_path
        return self.args.file_path

    ######################################################
    ## Write to file
    ######################################################

    def write_pandas_df(self, df: Any = None) -> bool:
        # File not yet initialized
        if self.args is None:
            return False

        # Check path is available
        if self.file_path is None:
            print_error("FilePath invalid")
            return False

        # write to file
        import pandas as pd

        if df is None or not isinstance(df, pd.DataFrame):
            print_error("DataFrame invalid")
            return False

        file_path: Path = self.file_path
        logger.debug("Writing DF to file: {}".format(file_path))
        if file_path.exists():
            logger.debug("Deleting: {}".format(file_path))
            delete_from_fs(file_path)

        if not file_path.exists():
            logger.debug("Creating: {}".format(file_path))
            file_path.parent.mkdir(parents=True, exist_ok=True)
            file_path.touch(exist_ok=True)

        try:
            with file_path.open("w") as open_file:
                logger.debug("file_type: {}".format(self.file_type))
                if self.file_type == FileType.JSON:
                    df.to_json(open_file, orient="index", indent=4)
                elif self.file_type == FileType.JSON:
                    df.to_csv(open_file)
                else:
                    print_error(f"{self.file_type.value} not yet supported")
                    return False
        except Exception as e:
            raise

        return True



from typing import Optional, Dict
from docker.models.containers import Container

from phidata.utils.log import logger


def execute_command(
    cmd: str,
    container: Container,
    wait: bool = True,
    print_output: bool = True,
    docker_env: Optional[Dict[str, str]] = None,
) -> bool:
    import os
    import socket
    import time

    logger.debug("Exec `{}` in container: {}".format(cmd, container.name))
    _container_socket: Optional[socket.SocketIO] = None
    exit_code, _container_socket = container.exec_run(
        cmd=cmd,
        stdout=True,
        stdin=True,
        tty=True,
        socket=True,
        environment=docker_env,
    )

    if _container_socket is None:
        logger.debug("container_socket is None")
        return False

    if not (wait or print_output):
        return True

    if _container_socket.readable():
        # code references the following
        # - https://stackoverflow.com/a/66329671
        # logger.debug("_container_socket is readable")
        _container_socket._sock.setblocking(False)  # type: ignore

        _op = None
        while True:
            try:
                _op = os.read(_container_socket.fileno(), 8192)
                # logger.debug("Op: {}".format(_op))
            except BlockingIOError as not_ready_error:
                # logger.exception(not_ready_error)
                logger.info("waiting...")
                time.sleep(10)
                continue

            if _op is None:
                logger.info("Output not available.")
                break
            if (
                _op == b""
                or (isinstance(_op, bytes) and _op.decode(errors="ignore")) == ""
            ):
                logger.info("Task finished")
                break

            # Use print here because console.print has issues decoding output
            if print_output:
                print(_op.decode(errors="ignore"))

    return True



from time import sleep
from typing import Any, Dict, List, Optional

from kubernetes.client import CustomObjectsApi
from kubernetes.client.models.v1_delete_options import V1DeleteOptions

from phidata.infra.k8s.api_client import K8sApiClient
from phidata.infra.k8s.resource.base import K8sResource
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class CustomObject(K8sResource):
    """
    The CustomResourceDefinition must be created before creating this object.
    When creating a CustomObject, provide the spec and generate the object body using
        get_k8s_object()

    References:
        * https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/CustomObjectsApi.md
        * https://github.com/kubernetes-client/python/blob/master/examples/custom_object.py
    """

    resource_type: str = "CustomObject"

    # CustomObject spec
    spec: Optional[Dict[str, Any]] = None

    # The custom resource's group name (required)
    group: str
    # The custom resource's version (required)
    version: str
    # The custom resource's plural name. For TPRs this would be lowercase plural kind. (required)
    plural: str

    # List of attributes to include in the K8s manifest
    fields_for_k8s_manifest: List[str] = ["spec"]

    def get_k8s_object(self) -> Dict[str, Any]:
        """Creates a body for this CustomObject"""

        _v1_custom_object = {
            "apiVersion": self.api_version,
            "kind": self.kind,
            "metadata": self.metadata.get_k8s_object().to_dict(),
            "spec": self.spec,
        }
        return _v1_custom_object

    @staticmethod
    def get_from_cluster(
        k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs
    ) -> Optional[List[Dict[str, Any]]]:
        """Reads CustomObject from K8s cluster.

        Args:
            k8s_client: K8sApiClient for the cluster
            namespace: Namespace to use.
        """
        if "group" not in kwargs:
            print_error("No Group provided when reading CustomObject")
            return None
        if "version" not in kwargs:
            print_error("No Version provided when reading CustomObject")
            return None
        if "plural" not in kwargs:
            print_error("No Plural provided when reading CustomObject")
            return None

        group = kwargs["group"]
        version = kwargs["version"]
        plural = kwargs["plural"]

        custom_objects_api: CustomObjectsApi = k8s_client.custom_objects_api
        custom_object_list: Optional[Dict[str, Any]] = None
        custom_objects: Optional[List[Dict[str, Any]]] = None
        try:
            if namespace:
                # logger.debug(
                #     f"Getting CustomObjects for:\n\tNS: {namespace}\n\tGroup: {group}\n\tVersion: {version}\n\tPlural: {plural}"
                # )
                custom_object_list = custom_objects_api.list_namespaced_custom_object(
                    group=group,
                    version=version,
                    namespace=namespace,
                    plural=plural,
                )
            else:
                # logger.debug(
                #     f"Getting CustomObjects for:\n\tGroup: {group}\n\tVersion: {version}\n\tPlural: {plural}"
                # )
                custom_object_list = custom_objects_api.list_cluster_custom_object(
                    group=group,
                    version=version,
                    plural=plural,
                )
        except Exception:
            logger.warning(f"Could not read custom objects for: {group}/{version}")
            logger.warning("Please check if the CustomResourceDefinition is created")
            return custom_objects

        # logger.debug(f"custom_object_list: {custom_object_list}")
        # logger.debug(f"custom_object_list type: {t    ype(custom_object_list)}")
        if custom_object_list:
            custom_objects = custom_object_list.get("items", None)
            # logger.debug(f"custom_objects: {custom_objects}")
            # logger.debug(f"custom_objects type: {type(custom_objects)}")
        return custom_objects

    def _create(self, k8s_client: K8sApiClient) -> bool:

        custom_objects_api: CustomObjectsApi = k8s_client.custom_objects_api
        k8s_object: Dict[str, Any] = self.get_k8s_object()
        namespace = self.get_namespace()

        print_info("Sleeping for 5 seconds so that CRDs can be registered")
        sleep(5)
        logger.debug("Creating: {}".format(self.get_resource_name()))
        custom_object: Dict[
            str, Any
        ] = custom_objects_api.create_namespaced_custom_object(
            group=self.group,
            version=self.version,
            namespace=namespace,
            plural=self.plural,
            body=k8s_object,
        )
        # logger.debug("Created:\n{}".format(pformat(custom_object, indent=2)))
        if custom_object.get("metadata", {}).get("creationTimestamp", None) is not None:
            logger.debug("CustomObject Created")
            self.active_resource = custom_object
            self.active_resource_class = Dict
            return True
        logger.error("CustomObject could not be created")
        return False

    def _read(self, k8s_client: K8sApiClient) -> Optional[Dict[str, Any]]:
        """Returns the "Active" CustomObject from the cluster"""

        namespace = self.get_namespace()
        active_resource: Optional[Dict[str, Any]] = None
        active_resources: Optional[List[Dict[str, Any]]] = self.get_from_cluster(
            k8s_client=k8s_client,
            namespace=namespace,
            group=self.group,
            version=self.version,
            plural=self.plural,
        )
        # logger.debug(f"active_resources: {active_resources}")
        if active_resources is None:
            return None

        active_resources_dict = {
            _custom_object.get("metadata", {}).get("name", None): _custom_object
            for _custom_object in active_resources
        }

        custom_object_name = self.get_resource_name()
        if custom_object_name in active_resources_dict:
            active_resource = active_resources_dict[custom_object_name]
            self.active_resource = active_resource
            self.active_resource_class = Dict
            logger.debug(f"Found active {custom_object_name}")
        return active_resource

    def _update(self, k8s_client: K8sApiClient) -> bool:

        custom_objects_api: CustomObjectsApi = k8s_client.custom_objects_api
        custom_object_name = self.get_resource_name()
        k8s_object: Dict[str, Any] = self.get_k8s_object()
        namespace = self.get_namespace()

        logger.debug("Updating: {}".format(custom_object_name))
        custom_object: Dict[
            str, Any
        ] = custom_objects_api.patch_namespaced_custom_object(
            group=self.group,
            version=self.version,
            namespace=namespace,
            plural=self.plural,
            name=custom_object_name,
            body=k8s_object,
        )
        # logger.debug("Updated: {}".format(custom_object))
        if custom_object.get("metadata", {}).get("creationTimestamp", None) is not None:
            logger.debug("CustomObject Updated")
            self.active_resource = custom_object
            self.active_resource_class = Dict
            return True
        logger.error("CustomObject could not be updated")
        return False

    def _delete(self, k8s_client: K8sApiClient) -> bool:

        custom_objects_api: CustomObjectsApi = k8s_client.custom_objects_api
        custom_object_name = self.get_resource_name()
        namespace = self.get_namespace()

        logger.debug("Deleting: {}".format(custom_object_name))
        self.active_resource = None
        delete_options = V1DeleteOptions()
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_status.py
        delete_status: Dict[
            str, Any
        ] = custom_objects_api.delete_namespaced_custom_object(
            group=self.group,
            version=self.version,
            namespace=namespace,
            plural=self.plural,
            name=custom_object_name,
            body=delete_options,
        )
        logger.debug("delete_status: {}".format(delete_status))
        if delete_status.get("status", None) == "Success":
            logger.debug("CustomObject Deleted")
            return True
        logger.error("CustomObject could not be deleted")
        return False



from typing import Optional, Any, Callable, Dict, List

from phidata.workflow import Workflow, WorkflowArgs
from phidata.utils.enums import ExtendedEnum
from phidata.utils.cli_console import print_error, print_info
from phidata.utils.log import logger


class EngineType(ExtendedEnum):
    PANDAS = "PANDAS"
    SPARK = "SPARK"
    DEFAULT = "DEFAULT"


class PythonWorkflowArgs(WorkflowArgs):
    # The compute engine for this python workflow
    # Pandas and Spark are the available options
    engine: EngineType = EngineType.DEFAULT
    # To run a workflow we need to execute a task.
    # For a PythonWorkflow this task is a python function
    # For example:
    #     def simple_python_function(**kwargs) -> bool:
    # The entrypoint() should accept **kwargs as the arguments
    # and return the run_status as a bool
    # The type for the entrypoint_function is Callable[..., bool].
    # https://stackoverflow.com/a/39624147
    # https://docs.python.org/3/library/typing.html#typing.Callable
    entrypoint: Callable[..., bool]
    entrypoint_args: Optional[List] = None
    entrypoint_kwargs: Optional[Dict] = None
    # If True, scans the args for SqlTable objects and populates the
    # SqlTable.db_engine value using the SqlTable.db_conn_id
    create_db_engine_from_conn_id: bool = False


class PythonWorkflow(Workflow):
    """Base Class for Python Workflows"""

    def __init__(self) -> None:
        super().__init__()
        self.args: Optional[PythonWorkflowArgs] = None

    @property
    def engine(self) -> EngineType:
        return self.args.engine if self.args else EngineType.DEFAULT

    @property
    def entrypoint(self) -> Optional[Callable[..., bool]]:
        return self.args.entrypoint if self.args else None

    ######################################################
    ## Run workflow
    ######################################################

    def run_in_local_env(self) -> bool:
        """
        Runs a python workflow in the local environment where phi wf is called from.

        Returns:
            run_status (bool): True if the run was successful
        """
        # Workflow not yet initialized
        if self.args is None:
            return False

        # Important: Validate that PathContext is available
        # This is derived from the current env or passed by the data product
        if self.path_context is None:
            print_error("PathContext not available")
            return False

        dry_run: bool = self.run_context.dry_run if self.run_context else False
        if dry_run:
            print_info("Dry run, returning True")
            return True

        run_status: bool = False
        entrypoint = self.entrypoint

        try:
            if self.args is not None and entrypoint is not None:
                run_status = entrypoint(**self.args.dict())
        except Exception as e:
            print_error("Could not run workflow")
            print_error(e)
        return run_status

    def run_in_docker_container(
        self, active_container: Any, docker_env: Optional[Dict[str, str]] = None
    ) -> bool:
        """
        Runs a python workflow in a docker container.

        Args:
            active_container: The container to run the workflow in
            docker_env:

        Returns:
            run_status (bool): True if the run was successful

        Notes:
            * This function runs in the local environment where phi wf is called from.
            But executes an `airflow` commands in the docker container to run the workflow
            * For the airflow tasks to be available, they need to be added to the workflow DAG
            using add_airflow_tasks_to_dag()
        """
        # Workflow not yet initialized
        if self.args is None:
            return False

        from pathlib import Path

        from docker.errors import APIError
        from docker.models.containers import Container

        from phidata.utils.cli_console import print_info, print_error
        from phidata.infra.docker.utils.container import execute_command

        if active_container is None or not isinstance(active_container, Container):
            print_error("Invalid Container object")
            return False
        container: Container = active_container

        task_id = self.task_id
        dag_id = self.dag_id
        if task_id is None:
            print_error("Workflow task_id unavailable")
            return False
        if dag_id is None:
            print_error("Workflow dag_id unavailable")
            return False

        run_date: Optional[str] = (
            self.run_context.run_date if self.run_context else None
        )
        dry_run: Optional[bool] = self.run_context.dry_run if self.run_context else None
        af_dry_run_str: str = " -n" if dry_run else ""
        workflow_file_path: Optional[Path] = (
            self.path_context.workflow_file if self.path_context else None
        )
        subdir_str: str = (
            f" -S {str(workflow_file_path.parent)}" if workflow_file_path else ""
        )

        # logger.debug("task_id: {}".format(task_id))
        # logger.debug("dag_id: {}".format(dag_id))
        # logger.debug("run_date: {}".format(run_date))
        # logger.debug("dry_run: {}".format(dry_run))

        run_status: bool = False
        list_tasks_cmd = f"airflow tasks list {dag_id} -t{subdir_str}"
        test_wf_cmd = f"airflow tasks test {dag_id} {task_id} {run_date}{subdir_str}{af_dry_run_str}"
        dry_run_cmd = f"python {workflow_file_path}"
        command_to_run = dry_run_cmd if dry_run else test_wf_cmd
        try:
            # execute_command(cmd=list_tasks, container=container)
            print_info(
                "Running command: {}\nContainer: {}".format(
                    command_to_run, container.name
                )
            )
            run_status = execute_command(
                cmd=command_to_run,
                container=container,
                docker_env=docker_env,
            )
        except APIError as e:
            logger.exception(e)
            raise

        return run_status

    def run_in_k8s_container(
        self,
        pod: Any,
        k8s_api_client: Any,
        container_name: Optional[str] = None,
        k8s_env: Optional[Dict[str, str]] = None,
    ) -> bool:
        """
        Runs a python workflow in a k8s container.

        Args:
            pod: The pod to run the workflow in
            k8s_api_client: K8sApiClient
            container_name: The container to run the workflow in
            k8s_env:

        Returns:
            run_status (bool): True if the run was successful

        Notes:
            * This function runs in the local environment where phi wf is called from.
            But executes an `airflow` commands in the k8s container to run the workflow
            * For the airflow tasks to be available, they need to be added to the workflow DAG
            using add_airflow_tasks_to_dag()
        """
        # Workflow not yet initialized
        if self.args is None:
            return False

        from pathlib import Path

        from phidata.infra.k8s.resource.core.v1.pod import Pod
        from phidata.infra.k8s.api_client import K8sApiClient

        from phidata.utils.cli_console import print_info, print_error
        from phidata.infra.k8s.utils.pod import execute_command

        if pod is None or not isinstance(pod, Pod):
            print_error("Invalid Pod")
            return False
        if k8s_api_client is None or not isinstance(k8s_api_client, K8sApiClient):
            print_error("Invalid K8sApiClient")
            return False

        task_id = self.task_id
        dag_id = self.dag_id
        if task_id is None:
            print_error("Workflow task_id unavailable")
            return False
        if dag_id is None:
            print_error("Workflow dag_id unavailable")
            return False

        run_date: Optional[str] = (
            self.run_context.run_date if self.run_context else None
        )
        dry_run: Optional[bool] = self.run_context.dry_run if self.run_context else None
        workflow_file_path: Optional[Path] = (
            self.path_context.workflow_file if self.path_context else None
        )
        subdir_str: str = (
            f"-S {str(workflow_file_path.parent)}" if workflow_file_path else ""
        )

        logger.debug("task_id: {}".format(task_id))
        logger.debug("dag_id: {}".format(dag_id))
        logger.debug("run_date: {}".format(run_date))
        logger.debug("dry_run: {}".format(dry_run))

        run_status: bool = False
        test_wf_cmd = ["airflow", "tasks", "test", dag_id, task_id, run_date]
        dry_run_cmd = ["python", workflow_file_path]
        command_to_run = dry_run_cmd if dry_run else test_wf_cmd
        try:
            run_status = execute_command(
                cmd=test_wf_cmd,
                pod=pod,
                k8s_api_client=k8s_api_client,
                container_name=container_name,
                k8s_env=k8s_env,
            )
        except Exception as e:
            print_error(e)
            raise

        return run_status

    ######################################################
    ## Airflow functions
    ######################################################

    def add_airflow_tasks_to_dag(self, dag: Any) -> bool:
        """
        This function adds the airflow tasks for this workflow to a DAG.
        This function is usually called by the create_airflow_dag()
        and runs on the remote machine where airflow is available.
        Args:
            dag:

        Returns:

        """
        # Workflow not yet initialized
        if self.args is None:
            return False

        # Important: Validate that PathContext is available
        if self.path_context is None:
            return False

        from airflow.models.dag import DAG
        from airflow.operators.python import PythonOperator

        if dag is None or not isinstance(dag, DAG):
            print_error("Invalid DAG")
            return False

        # Get the task_id for the airflow task
        task_id = self.task_id
        # logger.info(f"Creating airflow task: {task_id}")

        # Create db_engines if selected
        if self.args.create_db_engine_from_conn_id:
            self.create_db_engines()

        # Function to run
        entrypoint = (
            self.args.entrypoint
            if self.args and isinstance(self.args.entrypoint, Callable)
            else simple_python_function
        )
        # Airflow task
        task = PythonOperator(
            task_id=task_id,
            python_callable=entrypoint,
            op_kwargs=self.args.dict(),
        )
        dag.add_task(task)
        logger.debug(f"Airflow task created: {task_id}")
        return True

    def create_db_engines(self):
        from phidata.asset.table.sql import SqlTable, SqlType

        def create_db_engine_using_conn_id(t: SqlTable):
            # logger.info(f"Creating DbEngine for: {t.name}")
            if t.sql_type == SqlType.POSTGRES:
                from airflow.providers.postgres.hooks.postgres import PostgresHook

                pg_hook = PostgresHook(postgres_conn_id=t.db_conn_id)
                t.db_engine = pg_hook.get_sqlalchemy_engine()

        logger.debug("Creating DbEngines for SqlTables")
        # logger.info(f"args: {self.args}")
        for field_key, field_value in self.args.__dict__.items():
            if isinstance(field_value, SqlTable):
                create_db_engine_using_conn_id(field_value)

        if self.args.entrypoint_kwargs is not None and isinstance(
            self.args.entrypoint_kwargs, dict
        ):
            for kwargs_key, kwargs_value in self.args.entrypoint_kwargs.items():
                if isinstance(kwargs_value, SqlTable):
                    create_db_engine_using_conn_id(kwargs_value)
        if self.args.entrypoint_args is not None and isinstance(
            self.args.entrypoint_args, list
        ):
            for entrypoint_arg in self.args.entrypoint_args:
                if isinstance(entrypoint_arg, SqlTable):
                    create_db_engine_using_conn_id(entrypoint_arg)


def simple_python_function(**kwargs) -> bool:
    print_error("Invalid entrypoint function")
    print_error("Args: {}".format(kwargs))
    return True



from pathlib import Path
from typing import Optional

from phidata.asset.file import File
from phidata.asset import DataAsset, DataAssetArgs
from phidata.infra.aws.resource.s3.bucket import S3Bucket
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class S3ObjectArgs(DataAssetArgs):
    key: str
    bucket: S3Bucket
    update_ok: bool = True


class S3Object(DataAsset):
    def __init__(
        self,
        key: str,
        bucket: S3Bucket,
        update_ok: bool = False,
        version: Optional[str] = None,
        enabled: bool = True,
    ) -> None:

        super().__init__()
        try:
            self.args: S3ObjectArgs = S3ObjectArgs(
                key=key,
                bucket=bucket,
                name=key,
                update_ok=update_ok,
                version=version,
                enabled=enabled,
            )
        except Exception as e:
            print_error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def key(self) -> str:
        return self.args.key

    @key.setter
    def key(self, key: str) -> None:
        if key is not None:
            self.args.key = key

    @property
    def bucket(self) -> S3Bucket:
        return self.args.bucket

    @bucket.setter
    def bucket(self, bucket: S3Bucket) -> None:
        if bucket is not None:
            self.args.bucket = bucket

    ######################################################
    ## Write to s3
    ######################################################

    def upload_file(self, file: File) -> bool:

        ######################################################
        ## Validate
        ######################################################

        # S3Object not yet initialized
        if self.args is None:
            return False

        # Check file_path is available
        file_path: Optional[Path] = file.file_path
        if file_path is None or not isinstance(file_path, Path):
            print_error("FilePath invalid")
            return False

        bucket_name = self.bucket.name
        object_key = self.args.key
        print_info("Uploading")
        print_info(f"  File: {file_path}")
        print_info(f"  To  : {bucket_name}/{object_key}")

        ######################################################
        ## Upload
        ######################################################
        try:
            from botocore.exceptions import ClientError

            s3_bucket = self.bucket.read()
            # logger.info("Bucket: {}".format(s3_bucket))
            # logger.info("Bucket type: {}".format(type(s3_bucket)))

            if s3_bucket is None:
                print_error(
                    f"Could not reach bucket: {bucket_name}. Please confirm it exists."
                )
                return False

            # upload file to bucket
            s3_object_size = 0
            s3_object = s3_bucket.Object(key=object_key)
            try:
                s3_object.load()
                s3_object_size = s3_object.content_length
                # logger.info("s3_object_size: {}".format(s3_object_size))
            except ClientError as e:
                pass
            if s3_object_size >= 1 and not self.args.update_ok:
                logger.info(f"Object {bucket_name}/{object_key} already exists")
                return False
            # logger.info("s3_object type: {}".format(type(s3_object)))
            # logger.info("s3_object: {}".format(s3_object))

            logger.info("Uploading S3Object")
            s3_object.upload_file(Filename=str(file_path.resolve()))
            logger.info("S3Object uploaded")
            return True
        except Exception as e:
            print_error("Could not upload S3Object, please try again")
            print_info("--- stacktrace ---")
            logger.exception(e)
            print_info("--- stacktrace ---")
        return False

    ######################################################
    ## Read from s3
    ######################################################

    def download_file(self, file: File) -> bool:

        ######################################################
        ## Validate
        ######################################################

        # S3Object not yet initialized
        if self.args is None:
            return False

        # Check file_path is available
        file_path: Optional[Path] = file.file_path
        if file_path is None or not isinstance(file_path, Path):
            print_error("FilePath invalid")
            return False

        bucket_name = self.bucket.name
        object_key = self.args.key
        print_info("Download")
        print_info(f"  Object : {bucket_name}/{object_key}")
        print_info(f"  To     : {file_path}")

        ######################################################
        ## Download
        ######################################################
        try:
            from botocore.exceptions import ClientError

            s3_bucket = self.bucket.read()
            # logger.info("Bucket: {}".format(s3_bucket))
            # logger.info("Bucket type: {}".format(type(s3_bucket)))

            if s3_bucket is None:
                print_error(
                    f"Could not reach bucket: {bucket_name}. Please confirm it exists."
                )
                return False

            # download object to file
            s3_object_size = 0
            s3_object = s3_bucket.Object(key=object_key)
            try:
                s3_object.load()
                s3_object_size = s3_object.content_length
                logger.info("s3_object_size: {}".format(s3_object_size))
            except ClientError as e:
                pass
            if s3_object_size == 0:
                logger.info(f"Object {bucket_name}/{object_key} does not exist")
                return False
            logger.info("s3_object type: {}".format(type(s3_object)))
            logger.info("s3_object: {}".format(s3_object))

            logger.info("Download S3Object to File")
            s3_object.download_file(Filename=str(file_path.resolve()))
            logger.info("S3Object downloaded")
            return True
        except Exception as e:
            print_error("Could not download S3Object, please try again")
            print_info("--- stacktrace ---")
            logger.exception(e)
            print_info("--- stacktrace ---")
        return False



from typing import List, Optional

from kubernetes.client import AppsV1Api
from kubernetes.client.models.v1_deployment import V1Deployment
from kubernetes.client.models.v1_deployment_list import V1DeploymentList
from kubernetes.client.models.v1_deployment_spec import V1DeploymentSpec
from kubernetes.client.models.v1_status import V1Status
from pydantic import Field

from phidata.infra.k8s.api_client import K8sApiClient
from phidata.infra.k8s.resource.base import K8sResource, K8sObject
from phidata.infra.k8s.resource.apps.v1.deployment_strategy import DeploymentStrategy
from phidata.infra.k8s.resource.core.v1.pod_template_spec import PodTemplateSpec
from phidata.infra.k8s.resource.meta.v1.label_selector import LabelSelector
from phidata.utils.dttm import utc_now_str
from phidata.utils.log import logger


class DeploymentSpec(K8sObject):
    """
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#deploymentspec-v1-apps
    """

    resource_type: str = "DeploymentSpec"

    # Minimum number of seconds for which a newly created pod should be ready
    # without any of its container crashing, for it to be considered available.
    # Defaults to 0 (pod will be considered available as soon as it is ready)
    min_ready_seconds: Optional[int] = Field(None, alias="minReadySeconds")
    # Indicates that the deployment is paused.
    paused: Optional[bool] = None
    # The maximum time in seconds for a deployment to make progress before it is considered to be failed.
    # The deployment controller will continue to process failed deployments and a condition with a
    # ProgressDeadlineExceeded reason will be surfaced in the deployment status.
    # Note that progress will not be estimated during the time a deployment is paused.
    # Defaults to 600s.
    progress_deadline_seconds: Optional[int] = Field(
        None, alias="progressDeadlineSeconds"
    )
    replicas: int
    # The number of old ReplicaSets to retain to allow rollback.
    # This is a pointer to distinguish between explicit zero and not specified.
    # Defaults to 10.
    revision_history_limit: Optional[int] = Field(None, alias="revisionHistoryLimit")
    # The selector field defines how the Deployment finds which Pods to manage
    selector: LabelSelector
    strategy: Optional[DeploymentStrategy] = None
    template: PodTemplateSpec

    def get_k8s_object(self) -> V1DeploymentSpec:

        # Return a V1DeploymentSpec object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_deployment_spec.py
        _strategy = self.strategy.get_k8s_object() if self.strategy else None
        _v1_deployment_spec = V1DeploymentSpec(
            min_ready_seconds=self.min_ready_seconds,
            paused=self.paused,
            progress_deadline_seconds=self.progress_deadline_seconds,
            replicas=self.replicas,
            revision_history_limit=self.revision_history_limit,
            selector=self.selector.get_k8s_object(),
            strategy=_strategy,
            template=self.template.get_k8s_object(),
        )
        return _v1_deployment_spec


class Deployment(K8sResource):
    """
    Deployments are used to run containers.
    Containers are run in Pods or ReplicaSets, and Deployments manages those Pods or ReplicaSets.
    A Deployment provides declarative updates for Pods and ReplicaSets.

    References:
        * Docs:
            https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#deployment-v1-apps
            https://kubernetes.io/docs/concepts/workloads/controllers/deployment/
        * Type: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_deployment.py
    """

    resource_type: str = "Deployment"

    spec: DeploymentSpec
    # If True, adds `kubectl.kubernetes.io/restartedAt` annotation on update
    # so the deployment is restarted even without any data change
    restart_on_update: bool = True

    # List of attributes to include in the K8s manifest
    fields_for_k8s_manifest: List[str] = ["spec"]

    def get_k8s_object(self) -> V1Deployment:
        """Creates a body for this Deployment"""

        # Return a V1Deployment object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_deployment.py
        _v1_deployment = V1Deployment(
            api_version=self.api_version,
            kind=self.kind,
            metadata=self.metadata.get_k8s_object(),
            spec=self.spec.get_k8s_object(),
        )
        return _v1_deployment

    @staticmethod
    def get_from_cluster(
        k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs
    ) -> Optional[List[V1Deployment]]:
        """Reads Deployments from K8s cluster.

        Args:
            k8s_client: The K8sApiClient for the current Cluster
            namespace: Namespace to use.
        """
        apps_v1_api: AppsV1Api = k8s_client.apps_v1_api
        deploy_list: Optional[V1DeploymentList] = None
        if namespace:
            # logger.debug(f"Getting deploys for ns: {namespace}")
            deploy_list = apps_v1_api.list_namespaced_deployment(
                namespace=namespace, **kwargs
            )
        else:
            # logger.debug("Getting deploys for all namespaces")
            deploy_list = apps_v1_api.list_deployment_for_all_namespaces(**kwargs)

        deploys: Optional[List[V1Deployment]] = None
        if deploy_list:
            deploys = deploy_list.items
        # logger.debug(f"deploys: {deploys}")
        # logger.debug(f"deploys type: {type(deploys)}")
        return deploys

    def _create(self, k8s_client: K8sApiClient) -> bool:

        apps_v1_api: AppsV1Api = k8s_client.apps_v1_api
        k8s_object: V1Deployment = self.get_k8s_object()
        namespace = self.get_namespace()

        logger.debug("Creating: {}".format(self.get_resource_name()))
        v1_deployment: V1Deployment = apps_v1_api.create_namespaced_deployment(
            namespace=namespace,
            body=k8s_object,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("Created: {}".format(v1_deployment))
        if v1_deployment.metadata.creation_timestamp is not None:
            logger.debug("Deployment Created")
            self.active_resource = v1_deployment
            self.active_resource_class = V1Deployment
            return True
        logger.error("Deployment could not be created")
        return False

    def _read(self, k8s_client: K8sApiClient) -> Optional[V1Deployment]:
        """Returns the "Active" Deployment from the cluster"""

        namespace = self.get_namespace()
        active_resource: Optional[V1Deployment] = None
        active_resources: Optional[List[V1Deployment]] = self.get_from_cluster(
            k8s_client=k8s_client,
            namespace=namespace,
        )
        # logger.debug(f"Active Resources: {active_resources}")
        if active_resources is None:
            return None

        active_resources_dict = {
            _deploy.metadata.name: _deploy for _deploy in active_resources
        }

        deploy_name = self.get_resource_name()
        if deploy_name in active_resources_dict:
            active_resource = active_resources_dict[deploy_name]
            self.active_resource = active_resource
            logger.debug(f"Found active {deploy_name}")
        return active_resource

    def _update(self, k8s_client: K8sApiClient) -> bool:

        # update `spec.template.metadata` section
        # to add `kubectl.kubernetes.io/restartedAt` annotation
        # https://github.com/kubernetes-client/python/issues/1378#issuecomment-779323573
        if self.restart_on_update:
            if self.spec.template.metadata.annotations is None:
                self.spec.template.metadata.annotations = {}
            self.spec.template.metadata.annotations[
                "kubectl.kubernetes.io/restartedAt"
            ] = utc_now_str()
            logger.debug(f"annotations: {self.spec.template.metadata.annotations}")

        apps_v1_api: AppsV1Api = k8s_client.apps_v1_api
        deploy_name = self.get_resource_name()
        k8s_object: V1Deployment = self.get_k8s_object()
        namespace = self.get_namespace()

        logger.debug("Updating: {}".format(deploy_name))
        v1_deployment: V1Deployment = apps_v1_api.patch_namespaced_deployment(
            name=deploy_name,
            namespace=namespace,
            body=k8s_object,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("Updated: {}".format(v1_deployment))
        if v1_deployment.metadata.creation_timestamp is not None:
            logger.debug("Deployment Updated")
            self.active_resource = v1_deployment
            self.active_resource_class = V1Deployment
            return True
        logger.error("Deployment could not be updated")
        return False

    def _delete(self, k8s_client: K8sApiClient) -> bool:

        apps_v1_api: AppsV1Api = k8s_client.apps_v1_api
        deploy_name = self.get_resource_name()
        namespace = self.get_namespace()

        logger.debug("Deleting: {}".format(deploy_name))
        self.active_resource = None
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_status.py
        delete_status: V1Status = apps_v1_api.delete_namespaced_deployment(
            name=deploy_name,
            namespace=namespace,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        logger.debug("delete_status: {}".format(delete_status.status))
        if delete_status.status == "Success":
            logger.debug("Deployment Deleted")
            return True
        logger.error("Deployment could not be deleted")
        return False



from typing import Optional, Literal, Union, List

from phidata.asset.table.sql import SqlTable
from phidata.utils.log import logger
from phidata.workflow.py.python_workflow import (
    PythonWorkflow,
    PythonWorkflowArgs,
    EngineType,
)


class RunSqlQueryArgs(PythonWorkflowArgs):
    query: str
    sql_table: SqlTable
    show_sample_data: bool = False
    load_result_to: Optional[SqlTable] = None
    # How to behave if the load_result_to already exists.
    # fail: Raise a ValueError.
    # replace: Drop the table before inserting new values.
    # append: Insert new values to the existing table.
    if_exists: Optional[Literal["fail", "replace", "append"]] = None
    # Write DataFrame index as a column.
    # Uses index_label as the column name in the table.
    index: Optional[bool] = None
    # Column label for index column(s).
    # If None is given (default) and index is True, then the index names are used.
    # A sequence should be given if the DataFrame uses MultiIndex.
    index_label: Optional[Union[str, List[str]]] = None
    # Specify the number of rows in each batch to be written at a time.
    # By default, all rows will be written at once.
    chunksize: Optional[int] = None
    create_db_engine_from_conn_id: bool = True


class RunSqlQuery(PythonWorkflow):
    def __init__(
        self,
        query: str,
        sql_table: SqlTable,
        show_sample_data: bool = False,
        load_result_to: Optional[SqlTable] = None,
        if_exists: Optional[Literal["fail", "replace", "append"]] = None,
        index: Optional[bool] = None,
        index_label: Optional[Union[str, List[str]]] = None,
        chunksize: Optional[int] = None,
        engine: EngineType = EngineType.PANDAS,
        name: str = "run_sql_query",
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ):
        super().__init__()
        try:
            self.args: RunSqlQueryArgs = RunSqlQueryArgs(
                query=query,
                sql_table=sql_table,
                show_sample_data=show_sample_data,
                load_result_to=load_result_to,
                if_exists=if_exists,
                index=index,
                index_label=index_label,
                chunksize=chunksize,
                engine=engine,
                name=name,
                task_id=task_id,
                dag_id=dag_id,
                version=version,
                enabled=enabled,
                entrypoint=run_sql_query,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def query(self) -> Optional[str]:
        return self.args.query

    @query.setter
    def query(self, query: str) -> None:
        if query is not None:
            self.args.query = query

    @property
    def sql_table(self) -> Optional[SqlTable]:
        return self.args.sql_table

    @sql_table.setter
    def sql_table(self, sql_table: SqlTable) -> None:
        if sql_table is not None:
            self.args.sql_table = sql_table


def run_sql_query_pandas(args: RunSqlQueryArgs) -> bool:

    import pandas as pd

    query: str = args.query
    sql_table: SqlTable = args.sql_table

    df: Optional[pd.DataFrame] = sql_table.run_sql_query(query)
    if df is None:
        logger.error("Result DataFrame is empty")
        return False

    if args.show_sample_data:
        logger.info("Sample data:\n{}".format(df.head()))

    if args.load_result_to is not None:
        load_sql_table = args.load_result_to
        if not isinstance(load_sql_table, SqlTable):
            logger.error("load_result_to value not of type SqlTable")
            return False

        load_success = load_sql_table.write_pandas_df(
            df=df,
            if_exists=args.if_exists,
            index=args.index,
            index_label=args.index_label,
            chunksize=args.chunksize,
        )
        return load_success

    return True


def run_sql_query(**kwargs) -> bool:
    args: RunSqlQueryArgs = RunSqlQueryArgs(**kwargs)

    if args.engine in (EngineType.PANDAS, EngineType.DEFAULT):
        return run_sql_query_pandas(args)
    else:
        logger.error(f"EngineType: {args.engine} not yet supported")
        return False



from typing import Optional, Any
from typing_extensions import Literal

from botocore.exceptions import ClientError

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class S3Bucket(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#service-resource
    """

    resource_type = "s3.Bucket"
    service_name = "s3"

    # name of bucket
    name: str
    # The canned ACL to apply to the bucket.
    acl: Optional[
        Literal["private", "public-read", "public-read-write", "authenticated-read"]
    ] = None
    grant_full_control: Optional[str] = None
    grant_read: Optional[str] = None
    grant_read_ACP: Optional[str] = None
    grant_write: Optional[str] = None
    grant_write_ACP: Optional[str] = None
    object_lock_enabled_for_bucket: Optional[bool] = None
    object_ownership: Optional[
        Literal["BucketOwnerPreferred", "ObjectWriter", "BucketOwnerEnforced"]
    ] = None

    skip_delete = True

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the s3.Bucket

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            service_resource = self.get_service_resource(aws_client)

            ## Create Bucket
            # Bucket names are GLOBALLY unique!
            # AWS will give you the IllegalLocationConstraintException if you collide
            # with an already existing bucket and you've specified a region different than
            # the region of the already existing bucket. If you happen to guess the correct region of the
            # existing bucket it will give you the BucketAlreadyExists exception.
            bucket_configuration = {}
            if aws_client.aws_region != "us-east-1":
                bucket_configuration["LocationConstraint"] = aws_client.aws_region
            bucket = service_resource.create_bucket(
                Bucket=self.name,
                # CreateBucketConfiguration=bucket_configuration,
                ACL=self.acl if self.acl is not None else "",
                # GrantFullControl=self.grant_full_control,
                # GrantRead=self.grant_read,
                # GrantReadACP=self.grant_read_ACP,
                # GrantWrite=self.grant_write,
                # GrantWriteACP=self.grant_write_ACP,
                # ObjectLockEnabledForBucket=self.object_lock_enabled_for_bucket,
                # ObjectOwnership=self.object_ownership,
            )
            # logger.debug(f"Bucket: {bucket}")
            # logger.debug(f"Bucket type: {type(bucket)}")

            ## Validate Bucket creation
            bucket.load()
            creation_date = bucket.creation_date
            logger.debug(f"creation_date: {creation_date}")
            if creation_date is not None:
                print_info(f"Bucket created: {bucket.name}")
                self.active_resource = bucket
                self.active_resource_class = bucket.__class__
                return True
            logger.error("Bucket could not be created")
        except Exception as e:
            logger.exception(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:
        ## Wait for Bucket to be created
        if self.wait_for_creation:
            try:
                print_info("Waiting for Bucket to be created")
                waiter = self.get_service_client(aws_client).get_waiter("bucket_exists")
                waiter.wait(
                    Bucket=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                print_error(
                    f"Waiter Bucket Stack {self.name} failed, downstream actions might fail."
                )
                print_error(e)
                print_error("---+---")
        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the s3.Bucket

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.info(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            service_resource = self.get_service_resource(aws_client)
            bucket = service_resource.Bucket(name=self.name)

            bucket.load()
            creation_date = bucket.creation_date
            logger.debug(f"Bucket creation_date: {creation_date}")
            if creation_date is not None:
                logger.debug(f"Bucket found: {bucket.name}")
                self.active_resource = bucket
                self.active_resource_class = bucket.__class__
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
            pass
        except Exception as e:
            logger.exception(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the s3.Bucket

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            bucket = self.read(aws_client)
            # logger.debug(f"Bucket: {bucket}")
            # logger.debug(f"Bucket type: {type(bucket)}")
            self.active_resource = None
            self.active_resource_class = None
            bucket.delete()
            print_info(f"Bucket deleted: {bucket}")
            return True
        except Exception as e:
            logger.exception(e)
        return False



from typing import Optional, Any, List

from botocore.exceptions import ClientError

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class CloudFormationStack(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudformation.html#service-resource
    """

    resource_type = "CloudFormationStack"
    service_name = "cloudformation"

    # StackName: The name must be unique in the Region in which you are creating the stack.
    name: str
    # Location of file containing the template body.
    # The URL must point to a template (max size: 460,800 bytes) that's located in an
    # Amazon S3 bucket or a Systems Manager document.
    template_url: str
    # parameters: Optional[List[Dict[str, Union[str, bool]]]] = None
    # disable_rollback: Optional[bool] = None

    skip_delete = False

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the CloudFormationStack

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # Create CloudFormationStack
            service_resource = self.get_service_resource(aws_client)
            stack = service_resource.create_stack(
                StackName=self.name,
                TemplateURL=self.template_url,
            )
            # logger.debug(f"Stack: {stack}")
            # logger.debug(f"Stack type: {type(stack)}")

            # Validate Stack creation
            stack.load()
            creation_time = stack.creation_time
            logger.debug(f"creation_time: {creation_time}")
            if creation_time is not None:
                print_info(
                    f"{self.get_resource_type()}: {self.get_resource_name()} created"
                )
                self.active_resource = stack
                return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be created.")
            print_error("Please try again.")
            print_error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:
        # Wait for Stack to be created
        if self.wait_for_creation:
            try:
                print_info(
                    f"Waiting for {self.get_resource_type()} to be created, this may take a while"
                )
                waiter = self.get_service_client(aws_client).get_waiter(
                    "stack_create_complete"
                )
                waiter.wait(
                    StackName=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                print_error("Waiter failed.")
                print_error("Please try again.")
                print_error(e)
                return False
        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the CloudFormationStack

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            service_resource = self.get_service_resource(aws_client)
            stack = service_resource.Stack(name=self.name)

            stack.load()
            creation_time = stack.creation_time
            logger.debug(f"creation_time: {creation_time}")
            if creation_time is not None:
                logger.debug(f"Stack found: {stack.stack_name}")
                self.active_resource = stack
                self.active_resource_class = stack.__class__
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
            pass
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error("Please try again.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the CloudFormationStack

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            stack = self.read(aws_client)
            # logger.debug(f"Stack: {stack}")
            # logger.debug(f"Stack type: {type(stack)}")
            self.active_resource = None
            stack.delete()
            print_info(
                f"{self.get_resource_type()}: {self.get_resource_name()} deleted"
            )
            return True
        except Exception as e:
            logger.error(e)
        return False

    def post_delete(self, aws_client: AwsApiClient) -> bool:
        # Wait for Stack to be deleted
        if self.wait_for_deletion:
            try:
                print_info(
                    f"Waiting for {self.get_resource_type()} to be created, this may take a while"
                )
                waiter = self.get_service_client(aws_client).get_waiter(
                    "stack_delete_complete"
                )
                waiter.wait(
                    StackName=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
                return True
            except Exception as e:
                print_error("Waiter failed.")
                print_error("Please try again.")
                print_error(e)
        return True

    def get_stack_resource(
        self, aws_client: AwsApiClient, logical_id: str
    ) -> Optional[Any]:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudformation.html#CloudFormation.StackResource
        # logger.debug(f"Getting StackResource {logical_id} for {self.name}")
        try:
            service_resource = self.get_service_resource(aws_client)
            stack_resource = service_resource.StackResource(self.name, logical_id)
            return stack_resource
        except Exception as e:
            logger.error(e)
        return None

    def get_private_subnets(
        self, aws_client: Optional[AwsApiClient] = None
    ) -> Optional[List[str]]:
        try:
            client: AwsApiClient = (
                aws_client
                if aws_client is not None
                else AwsApiClient(
                    aws_region=self.get_aws_region(), aws_profile=self.get_aws_profile()
                )
            )

            private_subnet_1_stack_resource = self.get_stack_resource(
                client, "PrivateSubnet01"
            )
            private_subnet_1_physical_resource_id = (
                private_subnet_1_stack_resource.physical_resource_id
                if private_subnet_1_stack_resource is not None
                else None
            )
            # logger.debug(f"private_subnet_1: {private_subnet_1_physical_resource_id}")
            private_subnet_2_stack_resource = self.get_stack_resource(
                client, "PrivateSubnet02"
            )
            private_subnet_2_physical_resource_id = (
                private_subnet_2_stack_resource.physical_resource_id
                if private_subnet_2_stack_resource is not None
                else None
            )
            # logger.debug(f"private_subnet_2: {private_subnet_2_physical_resource_id}")

            private_subnets = []
            if private_subnet_1_physical_resource_id is not None:
                private_subnets.append(private_subnet_1_physical_resource_id)
            if private_subnet_2_physical_resource_id is not None:
                private_subnets.append(private_subnet_2_physical_resource_id)

            return private_subnets if (len(private_subnets) > 0) else None
        except Exception as e:
            logger.error(e)
        return None

    def get_public_subnets(
        self, aws_client: Optional[AwsApiClient] = None
    ) -> Optional[List[str]]:
        try:
            client: AwsApiClient = (
                aws_client
                if aws_client is not None
                else AwsApiClient(
                    aws_region=self.get_aws_region(), aws_profile=self.get_aws_profile()
                )
            )

            public_subnet_1_stack_resource = self.get_stack_resource(
                client, "PublicSubnet01"
            )
            public_subnet_1_physical_resource_id = (
                public_subnet_1_stack_resource.physical_resource_id
                if public_subnet_1_stack_resource is not None
                else None
            )
            # logger.debug(f"public_subnet_1: {public_subnet_1_physical_resource_id}")
            public_subnet_2_stack_resource = self.get_stack_resource(
                client, "PublicSubnet02"
            )
            public_subnet_2_physical_resource_id = (
                public_subnet_2_stack_resource.physical_resource_id
                if public_subnet_2_stack_resource is not None
                else None
            )
            # logger.debug(f"public_subnet_2: {public_subnet_2_physical_resource_id}")

            public_subnets = []
            if public_subnet_1_physical_resource_id is not None:
                public_subnets.append(public_subnet_1_physical_resource_id)
            if public_subnet_2_physical_resource_id is not None:
                public_subnets.append(public_subnet_2_physical_resource_id)

            return public_subnets if (len(public_subnets) > 0) else None
        except Exception as e:
            logger.error(e)
        return None

    def get_security_group(
        self, aws_client: Optional[AwsApiClient] = None
    ) -> Optional[str]:
        try:
            client: AwsApiClient = (
                aws_client
                if aws_client is not None
                else AwsApiClient(
                    aws_region=self.get_aws_region(), aws_profile=self.get_aws_profile()
                )
            )

            security_group_stack_resource = self.get_stack_resource(
                client, "ControlPlaneSecurityGroup"
            )
            security_group_physical_resource_id = (
                security_group_stack_resource.physical_resource_id
                if security_group_stack_resource is not None
                else None
            )
            logger.debug(f"security_group: {security_group_physical_resource_id}")

            return security_group_physical_resource_id
        except Exception as e:
            logger.error(e)
        return None



from pathlib import Path
from typing import Optional, Any

from pydantic import BaseModel

from phidata.constants import STORAGE_DIR_ENV_VAR
from phidata.utils.log import logger


class DataAssetArgs(BaseModel):
    name: Optional[str] = None
    version: Optional[str] = None
    enabled: bool = True

    # The base_dir_path is the root dir for the data_asset.
    # It depends on the environment (local vs container) and is
    # used by the data_asset to build the absolute path in different environments
    base_dir_path: Optional[Path] = None

    debug_on: bool = True

    class Config:
        arbitrary_types_allowed = True


class DataAsset:
    """Base Class for all DataAssets"""

    def __init__(self) -> None:
        self.args: Optional[DataAssetArgs] = None

    @property
    def name(self) -> str:
        return (
            self.args.name if self.args and self.args.name else self.__class__.__name__
        )

    @property
    def version(self) -> Optional[str]:
        return self.args.version if self.args else None

    @property
    def enabled(self) -> bool:
        return self.args.enabled if self.args else False

    @property
    def base_dir_path(self) -> Optional[Path]:
        """
        Returns the base dir for the data_asset.
        This base dir depends on the environment (local vs container) and is
        used by the data_asset to build the absolute path.
        """
        # data_asset not yet initialized
        if self.args is None:
            return None

        if self.args.base_dir_path:
            # use cached value if available
            return self.args.base_dir_path

        # logger.debug(f"Loading {STORAGE_DIR_ENV_VAR} from env")
        import os

        storage_dir_env = os.getenv(STORAGE_DIR_ENV_VAR)
        # logger.debug(f"{STORAGE_DIR_ENV_VAR}: {storage_dir_env}")
        if storage_dir_env is not None:
            self.args.base_dir_path = Path(storage_dir_env)
        return self.args.base_dir_path

    @base_dir_path.setter
    def base_dir_path(self, data_dir_path: Optional[Path]) -> None:
        if self.args is not None and data_dir_path is not None:
            self.args.base_dir_path = data_dir_path

    ######################################################
    ## Build and validate data asset
    ######################################################

    def build(self) -> bool:
        logger.debug(f"@build not defined for {self.__class__.__name__}")
        return False

    ######################################################
    ## Write to DataAsset
    ######################################################

    def write_pandas_df(self, df: Any = None) -> bool:
        logger.debug(f"@write_pandas_df not defined for {self.__class__.__name__}")
        return False

    ######################################################
    ## Print
    ######################################################

    def __str__(self) -> str:
        if self.args is not None:
            return self.args.json(indent=2, exclude_none=True, exclude={"enabled"})
        else:
            return self.__class__.__name__



from phidata.workflow.py.python_workflow import (
    PythonWorkflow,
    PythonWorkflowArgs,
    EngineType,
)
from phidata.workflow.py.python_function import PythonFunction
from phidata.workflow.py.create_workflow import create_workflow



from functools import wraps, partial

from phidata.workflow.py.python_function import PythonFunction
from phidata.utils.cli_console import print_error
from phidata.utils.log import logger


def create_workflow(func):
    """Converts a function into a python workflow"""

    wraps(func)

    def wrapper(*args, **kwargs):
        # logger.info(f"args: {args}")
        # logger.info(f"kwargs: {kwargs}")
        try:
            name = kwargs.get("name", func.__name__)
            entrypoint = partial(func, *args, **kwargs)

            workflow_kwargs = {}
            if "task_id" in kwargs:
                workflow_kwargs["task_id"] = kwargs["task_id"]
            if "dag_id" in kwargs:
                workflow_kwargs["dag_id"] = kwargs["dag_id"]
            if "version" in kwargs:
                workflow_kwargs["version"] = kwargs["version"]
            if "enabled" in kwargs:
                workflow_kwargs["enabled"] = kwargs["enabled"]
            if "engine" in kwargs:
                workflow_kwargs["engine"] = kwargs["engine"]
            workflow_kwargs["create_db_engine_from_conn_id"] = kwargs.get(
                "create_db_engine_from_conn_id", True
            )
            # logger.info(f"workflow_kwargs: {workflow_kwargs}")
            entrypoint_args_list = list(args) if args is not None else None
            entrypoint_kwargs = kwargs
            wf = PythonFunction(
                name=name,
                entrypoint=entrypoint,
                entrypoint_args=entrypoint_args_list,
                entrypoint_kwargs=entrypoint_kwargs,
                **workflow_kwargs
            )
            return wf
        except Exception as e:
            print_error("Could not create workflow for: {}".format(func))
            print_error(e)

    return wrapper


"""
How to use create_workflow:

@create_workflow
def test_wf(param, **kwargs):
    return True

wf = test_wf(param="yum")

dp = DataProduct(name="dp", workflows=[wf])
"""



from pathlib import Path
from typing import Optional, Dict
from typing_extensions import Literal

from phidata.infra.base.args import InfraArgs
from phidata.utils.log import logger


class InfraConfig:
    """Base Class for all phidata infra configs"""

    def __init__(self) -> None:
        self.args: Optional[InfraArgs] = None

    @property
    def name(self) -> str:
        if self.args and self.args.name:
            return self.args.name
        return self.__class__.__name__.lower()

    @property
    def env(self) -> Optional[Literal["dev", "stg", "prd"]]:
        return self.args.env if self.args else None

    @property
    def version(self) -> Optional[str]:
        return self.args.version if self.args else None

    @property
    def enabled(self) -> bool:
        return self.args.enabled if self.args else False

    @property
    def workspace_root_path(self) -> Optional[Path]:
        return self.args.workspace_root_path if self.args else None

    @workspace_root_path.setter
    def workspace_root_path(self, workspace_root_path: Path) -> None:
        if self.args is not None and workspace_root_path is not None:
            self.args.workspace_root_path = workspace_root_path

    @property
    def workspace_config_file_path(self) -> Optional[Path]:
        return self.args.workspace_config_file_path if self.args else None

    @workspace_config_file_path.setter
    def workspace_config_file_path(self, workspace_config_file_path: Path) -> None:
        if self.args is not None and workspace_config_file_path is not None:
            self.args.workspace_config_file_path = workspace_config_file_path

    @property
    def scripts_dir(self) -> Optional[str]:
        return self.args.scripts_dir if self.args else None

    @scripts_dir.setter
    def scripts_dir(self, scripts_dir: str) -> None:
        if self.args is not None and scripts_dir is not None:
            self.args.scripts_dir = scripts_dir

    @property
    def storage_dir(self) -> Optional[str]:
        return self.args.storage_dir if self.args else None

    @storage_dir.setter
    def storage_dir(self, storage_dir: str) -> None:
        if self.args is not None and storage_dir is not None:
            self.args.storage_dir = storage_dir

    @property
    def meta_dir(self) -> Optional[str]:
        return self.args.meta_dir if self.args else None

    @meta_dir.setter
    def meta_dir(self, meta_dir: str) -> None:
        if self.args is not None and meta_dir is not None:
            self.args.meta_dir = meta_dir

    @property
    def products_dir(self) -> Optional[str]:
        return self.args.products_dir if self.args else None

    @products_dir.setter
    def products_dir(self, products_dir: str) -> None:
        if self.args is not None and products_dir is not None:
            self.args.products_dir = products_dir

    @property
    def notebooks_dir(self) -> Optional[str]:
        return self.args.notebooks_dir if self.args else None

    @notebooks_dir.setter
    def notebooks_dir(self, notebooks_dir: str) -> None:
        if self.args is not None and notebooks_dir is not None:
            self.args.notebooks_dir = notebooks_dir

    @property
    def workspace_config_dir(self) -> Optional[str]:
        return self.args.workspace_config_dir if self.args else None

    @workspace_config_dir.setter
    def workspace_config_dir(self, workspace_config_dir: str) -> None:
        if self.args is not None and workspace_config_dir is not None:
            self.args.workspace_config_dir = workspace_config_dir

    @property
    def local_env(self) -> Optional[Dict[str, str]]:
        return self.args.local_env if self.args else None

    @local_env.setter
    def local_env(self, local_env: Dict[str, str]) -> None:
        if self.args is not None and local_env is not None:
            self.args.local_env = local_env

    @property
    def docker_env(self) -> Optional[Dict[str, str]]:
        return self.args.docker_env if self.args else None

    @docker_env.setter
    def docker_env(self, docker_env: Dict[str, str]) -> None:
        if self.args is not None and docker_env is not None:
            self.args.docker_env = docker_env

    @property
    def k8s_env(self) -> Optional[Dict[str, str]]:
        return self.args.k8s_env if self.args else None

    @k8s_env.setter
    def k8s_env(self, k8s_env: Dict[str, str]) -> None:
        if self.args is not None and k8s_env is not None:
            self.args.k8s_env = k8s_env

    def is_valid(self) -> bool:
        logger.debug(f"@is_valid not defined for {self.__class__.__name__}")
        # each subclass should implement this function
        return True



from typing import Dict, List, Optional

from pydantic import BaseModel

from phidata.infra.k8s.enums.api_group import ApiGroup
from phidata.infra.k8s.enums.api_version import ApiVersion
from phidata.infra.k8s.enums.kind import Kind
from phidata.infra.k8s.resource.rbac_authorization_k8s_io.v1.cluste_role_binding import (
    Subject,
    RoleRef,
    ClusterRoleBinding,
)
from phidata.infra.k8s.create.common.labels import create_component_labels_dict
from phidata.infra.k8s.resource.meta.v1.object_meta import ObjectMeta
from phidata.utils.cli_console import print_info
from phidata.utils.log import logger


class CreateClusterRoleBinding(BaseModel):
    crb_name: str
    cr_name: str
    service_account_name: str
    app_name: str
    namespace: str
    labels: Optional[Dict[str, str]] = None

    def create(self) -> ClusterRoleBinding:
        """Creates the ClusterRoleBinding resource"""

        crb_name = self.crb_name
        logger.debug(f"Init ClusterRoleBinding resource: {crb_name}")

        sa_name = self.service_account_name
        subjects: List[Subject] = [
            Subject(kind=Kind.SERVICEACCOUNT, name=sa_name, namespace=self.namespace)
        ]
        cr_name = self.cr_name

        crb_labels = create_component_labels_dict(
            component_name=crb_name,
            app_name=self.app_name,
            labels=self.labels,
        )
        crb = ClusterRoleBinding(
            api_version=ApiVersion.RBAC_AUTH_V1,
            kind=Kind.CLUSTERROLEBINDING,
            metadata=ObjectMeta(
                name=crb_name,
                labels=crb_labels,
            ),
            role_ref=RoleRef(
                api_group=ApiGroup.RBAC_AUTH,
                kind=Kind.CLUSTERROLE,
                name=cr_name,
            ),
            subjects=subjects,
        )
        return crb
